# -*- coding: utf-8 -*-
"""VEM BOLD Constrained

File that contains function for BOLD data analysis with positivity
and l2-norm=1 constraints.

It imports functions from vem_tools.py in pyhrf/vbjde
"""

import time
import copy
import logging
import os

import os.path as op
import numpy as np

import pyhrf
import pyhrf.vbjde.vem_tools as vt

from pyhrf.boldsynth.hrf import getCanoHRF, genGaussianSmoothHRF
from pyhrf.sandbox.physio_params import (PHY_PARAMS_KHALIDOV11,
                                         linear_rf_operator,
                                         create_physio_brf,
                                         create_physio_prf)

import matplotlib
import matplotlib.pyplot as plt
try:
    os.environ["DISPLAY"]
except KeyError:
    matplotlib.use("Agg")
    plt.switch_backend("Agg")
else:
    try:
        matplotlib.use("Qt4Agg")
        plt.switch_backend("Qt4Agg")
    except ImportError:
        pass


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

eps = 1e-6

#@profile
def Main_vbjde_physio(graph, Y, Onsets, durations, Thrf, K, TR, beta, dt,
                      scale=1, estimateSigmaH=True, estimateSigmaG=True,
                      sigmaH=0.05, sigmaG=0.05, gamma_h=0, gamma_g=0,
                      NitMax=-1, NitMin=1, estimateBeta=True, PLOT=False,
                      contrasts=[], computeContrast=False,
                      idx_first_tag=0, simulation=None, sigmaMu=None,
                      estimateH=True, estimateG=True, estimateA=True,
                      estimateC=True, estimateZ=True, estimateNoise=True,
                      estimateMP=True, estimateLA=True, use_hyperprior=False,
                      positivity=False, constraint=False,
                      phy_params=PHY_PARAMS_KHALIDOV11, prior='omega', zc=False):

    logger.info("EM for ASL!")
    np.random.seed(6537540)
    logger.info("data shape: ")
    logger.info(Y.shape)

    Thresh = 1e-5
    D, M = np.int(np.ceil(Thrf / dt)) + 1, len(Onsets)
    #D, M = np.int(np.ceil(Thrf / dt)), len(Onsets)
    n_sess, N, J = Y.shape[0], Y.shape[1], Y.shape[2]

    Crit_AH, Crit_CG, cTime, rerror, FE = 1, 1, [], [], []
    EP, EPlh, Ent = [],[],[]
    Crit_H, Crit_G, Crit_Z, Crit_A, Crit_C = 1, 1, 1, 1, 1
    cAH, cCG, AH1, CG1 = [], [], [], []
    cA, cC, cH, cG, cZ = [], [], [], [], []
    h_norm, g_norm = [], []
    SUM_q_Z = [[] for m in xrange(M)]
    mua1 = [[] for m in xrange(M)]
    muc1 = [[] for m in xrange(M)]
    sigmaH = sigmaH * J / 100
    print sigmaH
    gamma_h = gamma_h * 100 / J
    print gamma_h

    # Beta data
    MaxItGrad = 200
    gradientStep = 0.005
    gamma = 7.5
    print 'gamma = ', gamma
    print 'voxels = ', J
    maxNeighbours, neighboursIndexes = vt.create_neighbours(graph, J)
    print 'graph.shape = ', graph.shape

    # Control-tag
    w = np.ones((N))
    w[idx_first_tag + 1::2] = -1
    W = np.diag(w)

    # Conditions
    print 'Onsets: ', Onsets
    print 'durations = ', durations
    print 'creating conditions...'
    X, XX, condition_names = vt.create_conditions_block_ms(Onsets, durations,
                                                    M, N, D, n_sess, TR, dt)
    # Covariance matrix
    #R = vt.covariance_matrix(2, D, dt)
    _, R_inv = genGaussianSmoothHRF(zc, D, dt, 1., 2)
    R = np.linalg.inv(R_inv)
    if zc:
        XX = XX[:, :, :, 1:-1]    # XX shape (S, M, N, D)
        D = D - 2
    AH1, CG1 = np.zeros((J, M, D)), np.zeros((J, M, D))

    print 'HRF length = ', D
    print 'Condition number = ', M
    print 'Number of scans = ', N
    print 'Number of voxels = ', J
    print 'Number of sessions = ', n_sess
    print 'XX.shape = ', XX.shape

    # Noise matrix
    Gamma = np.identity(N)
    # Noise initialization
    sigma_eps = np.ones((n_sess, J))
    # Labels
    logger.info("Labels are initialized by setting active probabilities "
                "to ones ...")
    q_Z = np.ones((M, K, J), dtype=np.float64) / 2.
    #q_Z = np.zeros((M, K, J), dtype=np.float64)
    #q_Z[:, 1, :] = 1
    q_Z1 = copy.deepcopy(q_Z)
    Z_tilde = copy.deepcopy(q_Z)

    # H and G
    TT, m_h = getCanoHRF(Thrf, dt)
    H = np.array(m_h[:D]).astype(np.float64)
    H /= np.linalg.norm(H)
    G = copy.deepcopy(H)
    Hb = create_physio_brf(phy_params, response_dt=dt, response_duration=Thrf)
    Hb /= np.linalg.norm(Hb)
    Gb = create_physio_prf(phy_params, response_dt=dt, response_duration=Thrf)
    Gb /= np.linalg.norm(Gb)
    if prior=='balloon':
        H = Hb.copy()
        G = Gb.copy()
    Mu = Hb.copy()
    H1 = copy.deepcopy(H)
    Sigma_H = np.zeros((D, D), dtype=np.float64)
    G1 = copy.deepcopy(G)
    Sigma_G = copy.deepcopy(Sigma_H)
    normOh = False
    normg = False
    if prior=='hierarchical' or prior=='omega':
        Omega = linear_rf_operator(len(H), phy_params, dt, calculating_brf=False)
    if prior=='omega':
        Omega0 = Omega.copy()
        OmegaH = np.dot(Omega, H)
        G = np.dot(Omega, H)
        if normOh or normg:
            Omega /= np.linalg.norm(OmegaH)
            OmegaH /=np.linalg.norm(OmegaH)
            G /= np.linalg.norm(G)

    # Initialize model parameters
    Beta = beta * np.ones((M), dtype=np.float64)
    n_drift = 4
    P = np.zeros((n_sess, N, n_drift+1), dtype=np.float64)
    L = np.zeros((n_drift+1, J, n_sess), dtype=np.float64)
    WP = np.zeros((n_sess, N, n_drift+2), dtype=np.float64)
    AL = np.zeros((n_drift+2, J, n_sess), dtype=np.float64)
    alpha = np.zeros((J), dtype=np.float64)
    for s in xrange(0, n_sess):
        P[s, :, :] = vt.PolyMat(N, n_drift, TR)
        L[:, :, s] = vt.polyFit(Y[s, :, :], TR, n_drift, P[s, :, :])
        WP[s, :, :] = np.append(w[:, np.newaxis], P[s, :, :], axis=1)
        AL[:, :, s] = np.append(alpha[np.newaxis, :], L[:, :, s], axis=0)
    PL = np.einsum('ijk,kli->ijl', WP, AL)
    y_tilde = Y - PL

    # Parameters Gaussian mixtures
    mu_Ma = np.append(np.zeros((M, 1)), np.ones((M, 1)), axis=1).astype(np.float64)
    mu_Mc = mu_Ma.copy()
    sigma_Ma = np.ones((M, K), dtype=np.float64) * 0.3
    sigma_Mc = sigma_Ma.copy()

    # Params RLs
    m_A = np.zeros((n_sess, J, M), dtype=np.float64)
    for s in xrange(0, n_sess):
        for j in xrange(0, J):
            m_A[s, j, :] = (np.random.normal(mu_Ma, np.sqrt(sigma_Ma)) * q_Z[:, :, j]).sum(axis=1).T
    m_A1 = m_A.copy()
    Sigma_A = np.ones((M, M, J, n_sess)) * np.identity(M)[:, :, np.newaxis, np.newaxis]
    m_C = m_A.copy()
    m_C1 = m_C.copy()
    Sigma_C = Sigma_A.copy()

    # Precomputations
    print 'W shape is ', W.shape
    WX = W.dot(XX).transpose(1, 2, 0, 3)                                       # shape (S, M, N, D)
    Gamma_X = np.zeros((N, n_sess, M, D), dtype=np.float64)                    # shape (N, S, M, D)
    X_Gamma_X = np.zeros((D, M, n_sess, M, D), dtype=np.float64)               # shape (D, M, S, M, D)
    Gamma_WX = np.zeros((N, n_sess, M, D), dtype=np.float64)                   # shape (N, S, M, D)
    XW_Gamma_WX = np.zeros((D, M, n_sess, M, D), dtype=np.float64)             # shape (D, M, S, M, D)
    Gamma_WP = np.zeros((N, n_sess, n_drift+2), dtype=np.float64)              # shape (N, S, M, D)
    WP_Gamma_WP = np.zeros((n_sess, n_drift+2, n_drift+2), dtype=np.float64)   # shape (D, M, S, M, D)
    for s in xrange(0, n_sess):
        Gamma_X[:, s, :, :] = np.tensordot(Gamma, XX[s, :, :, :], axes=(1, 1))
        X_Gamma_X[:, :, s, :, :] = np.tensordot(XX[s, :, :, :].T, Gamma_X[:, s, :, :], axes=(1, 0))
        Gamma_WX[:, s, :, :] = np.tensordot(Gamma, WX[s, :, :, :], axes=(1, 1))
        XW_Gamma_WX[:, :, s, :, :] = np.tensordot(WX[s, :, :, :].T, Gamma_WX[:, s, :, :], axes=(1, 0))
        Gamma_WP[:, s, :] = Gamma.dot(WP[s, :, :])                             # (N, n_drift)
        WP_Gamma_WP[s, :, :] = WP[s, :, :].T.dot(Gamma_WP[:, s, :])            # (n_drift, n_drift)
    sigma_eps_m = np.maximum(sigma_eps, eps)                                   # (n_sess, J)
    cov_noise = sigma_eps_m[:, :, np.newaxis, np.newaxis]                      # (n_sess, J, 1, 1)


    ###########################################################################
    #############################################             VBJDE

    t1 = time.time()
    ni = 0

    #while ((ni < NitMin + 1) or (((Crit_AH > Thresh) or (Crit_CG > Thresh)) \
    #        and (ni < NitMax))):
    #while ((ni < NitMin + 1) or (((Crit_AH > Thresh)) \
    #        and (ni < NitMax))):
    while ((ni < NitMin + 1) or (((Crit_FE > Thresh * np.ones_like(Crit_FE)).any()) \
            and (ni < NitMax))):

        logger.info("-------- Iteration n° " + str(ni + 1) + " --------")

        if PLOT and ni >= 0:  # Plotting HRF and PRF
            logger.info("Plotting HRF and PRF for current iteration")
            vt.plot_response_functions_it(ni, NitMin, M, H, G)


        # Managing types of prior
        priorH_cov_term = np.zeros_like(R_inv)
        priorG_cov_term = np.zeros_like(R_inv)
        matrix_covH = R_inv.copy()
        matrix_covG = R_inv.copy()
        if prior=='balloon':
            logger.info("   prior balloon")
            #matrix_covH = np.eye(R_inv.shape[0], R_inv.shape[1])
            #matrix_covG = np.eye(R_inv.shape[0], R_inv.shape[1])
            priorH_mean_term = np.dot(matrix_covH / sigmaH, Hb)
            priorG_mean_term = np.dot(matrix_covG / sigmaG, Gb)
        elif prior=='omega':
            logger.info("   prior omega")
            #matrix_covG = np.eye(R_inv.shape[0], R_inv.shape[1])
            priorH_mean_term = np.dot(np.dot(Omega.T, matrix_covG / sigmaG), G)
            priorH_cov_term = np.dot(np.dot(Omega.T, matrix_covG / sigmaG), Omega)
            priorG_mean_term = np.dot(matrix_covG / sigmaG, OmegaH)
        elif prior=='hierarchical':
            logger.info("   prior hierarchical")
            matrix_covH = np.eye(R_inv.shape[0], R_inv.shape[1])
            matrix_covG = np.eye(R_inv.shape[0], R_inv.shape[1])
            priorH_mean_term = Mu / sigmaH
            priorG_mean_term = np.dot(Omega, Mu / sigmaG)
        else:
            logger.info("   NO prior")
            priorH_mean_term = np.zeros_like(H)
            priorG_mean_term = np.zeros_like(G)



        #####################
        # EXPECTATION
        #####################


        # HRF H
        if estimateH:
            logger.info("E H step ...")
            Ht, Sigma_H = vt.expectation_H_ms(Sigma_A, m_A, m_C, G, XX, W, Gamma,
                                            Gamma_X, X_Gamma_X, J, y_tilde,
                                            cov_noise, matrix_covH, sigmaH,
                                            priorH_mean_term, priorH_cov_term, N, M, D, n_sess)

            if constraint:
                if not np.linalg.norm(Ht)==1:
                    logger.info("   constraint l2-norm = 1")
                    H = vt.constraint_norm1_b(Ht, Sigma_H)
                    #H = Ht / np.linalg.norm(Ht)
                else:
                    logger.info("   l2-norm already 1!!!!!")
                    H = Ht.copy()
                Sigma_H = np.zeros_like(Sigma_H)
            else:
                H = Ht.copy()
                h_norm = np.append(h_norm, np.linalg.norm(H))
                print 'h_norm = ', h_norm

            Crit_H = (np.linalg.norm(H - H1) / np.linalg.norm(H1)) ** 2
            cH += [Crit_H]
            H1[:] = H[:]
            if prior=='omega':
                OmegaH = np.dot(Omega0, H)
                Omega = Omega0
                if normOh:
                    Omega /= np.linalg.norm(OmegaH)
                    OmegaH /= np.linalg.norm(OmegaH)


        # A
        if estimateA:
            logger.info("E A step ...")
            m_A, Sigma_A = vt.expectation_A_ms(m_A, Sigma_A, H, G, m_C, W, XX,
                                             Gamma, Gamma_X, q_Z,
                                             mu_Ma, sigma_Ma, J, y_tilde,
                                             Sigma_H, sigma_eps_m, N, M, D, n_sess)

            cA += [(np.linalg.norm(m_A - m_A1) / np.linalg.norm(m_A1)) ** 2]
            m_A1[:, :, :] = m_A[:, :, :]


        # PRF G
        if estimateG:
            logger.info("E G step ...")
            Gt, Sigma_G = vt.expectation_G_ms(Sigma_C, m_C, m_A, H, XX, W, WX, Gamma,
                                            Gamma_WX, XW_Gamma_WX, J, y_tilde,
                                            cov_noise, matrix_covG, sigmaG,
                                            priorG_mean_term, priorG_cov_term, N, M, D, n_sess)

            if constraint and normg:
                if not np.linalg.norm(Gt)==1:
                    logger.info("   constraint l2-norm = 1")
                    G = vt.constraint_norm1_b(Gt, Sigma_G, positivity=positivity)
                    #G = Gt / np.linalg.norm(Gt)
                else:
                    logger.info("   l2-norm already 1!!!!!")
                    G = Gt.copy()
                Sigma_G = np.zeros_like(Sigma_G)
            else:
                G = Gt.copy()
                g_norm = np.append(g_norm, np.linalg.norm(G))
                print 'g_norm = ', g_norm
            cG += [(np.linalg.norm(G - G1) / np.linalg.norm(G1)) ** 2]
            G1[:] = G[:]

        # C
        if estimateC:
            logger.info("E C step ...")
            m_C, Sigma_C = vt.expectation_C_ms(m_C, Sigma_C, G, H, m_A, W, XX, Gamma, Gamma_X, q_Z,
                                             mu_Mc, sigma_Mc, J, y_tilde,
                                             Sigma_G, sigma_eps_m, N, M, D, n_sess)

            cC += [(np.linalg.norm(m_C - m_C1) / np.linalg.norm(m_C1)) ** 2]
            m_C1[:, :] = m_C[:, :]


        # Q labels
        if estimateZ:
            logger.info("E Q step ...")
            q_Z, Z_tilde = vt.expectation_Q_ms(Sigma_A, m_A, Sigma_C, m_C,
                                            sigma_Ma, mu_Ma, sigma_Mc, mu_Mc,
                                            Beta, Z_tilde, q_Z, neighboursIndexes, graph, M, J, K, n_sess)

            if 0:
                import matplotlib.pyplot as plt
                plt.close('all')
                fig = plt.figure(1)
                for m in xrange(M):
                    ax = fig.add_subplot(2, M, m + 1)
                    im = ax.matshow(m_A[:, :, m].mean(0).reshape(20, 20))
                    plt.colorbar(im, ax=ax)
                    ax = fig.add_subplot(2, M, m + 3)
                    im = ax.matshow(q_Z[m, 1, :].reshape(20, 20))
                    plt.colorbar(im, ax=ax)
                fig = plt.figure(2)
                for m in xrange(M):
                    for s in xrange(n_sess):
                        ax = fig.add_subplot(M, n_sess, n_sess * m + s + 1)
                        im = ax.matshow(m_A[s, :, m].reshape(20, 20))
                        plt.colorbar(im, ax=ax)
                plt.show()

            cZ += [(np.linalg.norm(q_Z - q_Z1) / (np.linalg.norm(q_Z1) + eps)) ** 2]
            q_Z1 = q_Z

        if ni > 0:
            free_energyE = 0
            for s in xrange(n_sess):
                free_energyE += vt.Compute_FreeEnergy(y_tilde[s, :, :], m_A[s, :, :], Sigma_A[:, :, :, s],
                                             mu_Ma, sigma_Ma, H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C[s, :, :], Sigma_C[:, :, :, s], mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps[s, :], XX[s, :, :, :],
                                             W, J, D, M, N, K, use_hyperprior, Gamma_X[:, s, :, :],
                                             Gamma_WX[:, s, :, :], bold=True, S=n_sess)
            if free_energyE < free_energy:
                logger.info("free energy has decreased after E step from %f to %f", free_energy, free_energyE)


        # crit. AH and CG
        logger.info("crit. AH and CG")
        AH = m_A[:, :, :, np.newaxis] * H[np.newaxis, np.newaxis, :]
        CG = m_C[:, :, :, np.newaxis] * G[np.newaxis, np.newaxis, :]

        Crit_AH = (np.linalg.norm(AH - AH1) / (np.linalg.norm(AH1) + eps)) ** 2
        cAH += [Crit_AH]
        AH1 = AH.copy()
        Crit_CG = (np.linalg.norm(CG - CG1) / (np.linalg.norm(CG1) + eps)) ** 2
        cCG += [Crit_CG]
        CG1 = CG.copy()
        logger.info("Crit_AH = " + str(Crit_AH))
        logger.info("Crit_CG = " + str(Crit_CG))


        #####################
        # MAXIMIZATION
        #####################

        if prior=='balloon':
            logger.info("   prior balloon")
            AuxH = H - Hb
            AuxG = G - Gb
        elif prior=='omega':
            logger.info("   prior omega")
            AuxH = H.copy()
            AuxG = G - np.dot(Omega, H) #/np.linalg.norm(np.dot(Omega, H))
        elif prior=='hierarchical':
            logger.info("   prior hierarchical")
            AuxH = H - Mu
            AuxG = G - np.dot(Omega, Mu)
        else:
            logger.info("   NO prior")
            AuxH = H.copy()
            AuxG = G.copy()

        # Variance HRF: sigmaH
        if estimateSigmaH:
            logger.info("M sigma_H step ...")
            sigmaH = vt.maximization_sigma_asl(D, Sigma_H, matrix_covH, AuxH, use_hyperprior, gamma_h)
            logger.info('sigmaH = ' + str(sigmaH))

        if ni > 0:
            free_energyVh = 0
            for s in xrange(n_sess):
                free_energyVh += vt.Compute_FreeEnergy(y_tilde[s, :, :], m_A[s, :, :], Sigma_A[:, :, :, s], mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C[s, :, :], Sigma_C[:, :, :, s], mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps[s, :], XX[s, :, :, :], W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X[:, s, :, :], Gamma_WX[:, s, :, :], bold=True, S=n_sess)
            if free_energyVh < free_energyE:
                logger.info("free energy has decreased after v_h computation from %f to %f", free_energyE, free_energyVh)

         # Variance PRF: sigmaG
        if estimateSigmaG:
            logger.info("M sigma_G step ...")
            sigmaG = vt.maximization_sigma_asl(D, Sigma_G, matrix_covG, AuxG, use_hyperprior, gamma_g)
            logger.info('sigmaG = ' + str(sigmaG))

        # Mu: True HRF in the hierarchical prior case
        if prior=='hierarchical':
            logger.info("M sigma_G step ...")
            Mu = vt.maximization_Mu(H, G, matrix_covH, matrix_covG,
                                     sigmaH, sigmaG, sigmaMu, Omega, R_inv)
            logger.info('sigmaG = ' + str(sigmaG))

        # (mu,sigma)
        if estimateMP:
            logger.info("M (mu,sigma) a and c step ...")
            mu_Ma, sigma_Ma = vt.maximization_mu_sigma_ms(q_Z, m_A, Sigma_A, M, J, n_sess, K)
            mu_Mc, sigma_Mc = vt.maximization_mu_sigma_ms(q_Z, m_C, Sigma_C, M, J, n_sess, K)

        if ni > 0:
            free_energyMP = 0
            for s in xrange(n_sess):
                free_energyMP += vt.Compute_FreeEnergy(y_tilde[s, :, :], m_A[s, :, :], Sigma_A[:, :, :, s], mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C[s, :, :], Sigma_C[:, :, :, s], mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps[s, :], XX[s, :, :, :], W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X[:, s, :, :], Gamma_WX[:, s, :, :], bold=True, S=n_sess)
            if free_energyMP < free_energyVh:
                logger.info("free energy has decreased after GMM parameters computation from %f to %f", free_energyVh, free_energyMP)


        # Drift L, alpha
        if estimateLA:
            logger.info("M L, alpha step ...")
            for s in xrange(n_sess):
                AL[:, :, s] = vt.maximization_LA_asl(Y[s, :, :], m_A[s, :, :], m_C[s, :, :], XX[s, :, :, :],
                                                     WP[s, :, :], W, WP_Gamma_WP[s, :, :], H, G, Gamma)
            PL = np.einsum('ijk,kli->ijl', WP, AL)
            y_tilde = Y - PL

        if ni > 0:
            free_energyLA = 0
            for s in xrange(n_sess):
                free_energyLA += vt.Compute_FreeEnergy(y_tilde[s, :, :], m_A[s, :, :], Sigma_A[:, :, :, s], mu_Ma, sigma_Ma,
                                                 H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                                 m_C[s, :, :], Sigma_C[:, :, :, s], mu_Mc, sigma_Mc, G, Sigma_G,
                                                 AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                                 gamma, gamma_h, gamma_g, sigma_eps[s, :], XX[s, :, :, :], W,
                                                 J, D, M, N, K, use_hyperprior, Gamma_X[:, s, :, :], Gamma_WX[:, s, :, :], bold=True, S=n_sess)
            if free_energyLA < free_energyMP:
                logger.info("free energy has decreased after drifts computation from %f to %f", free_energyMP, free_energyLA)


        # Beta
        if estimateBeta:
            logger.info("M beta step ...")
            """Qtilde = np.concatenate((Z_tilde, np.zeros((M, K, 1), dtype=Z_tilde.dtype)), axis=2)
            Qtilde_sumneighbour = Qtilde[:, :, neighboursIndexes].sum(axis=3)
            Beta = vt.maximization_beta_m2(Beta.copy(), q_Z, Qtilde_sumneighbour,
                                             Qtilde, neighboursIndexes, maxNeighbours,
                                             gamma, MaxItGrad, gradientStep)
            logger.info(Beta)
            """
            Qtilde = np.concatenate((Z_tilde, np.zeros((M, K, 1), dtype=Z_tilde.dtype)), axis=2)
            Qtilde_sumneighbour = Qtilde[:, :, neighboursIndexes].sum(axis=3)
            for m in xrange(0, M):
                Beta[m] = vt.maximization_beta_m2_scipy_asl(Beta[m].copy(), q_Z[m, :, :], Qtilde_sumneighbour[m, :, :],
                                                   Qtilde[m, :, :], neighboursIndexes, maxNeighbours,
                                                   gamma, MaxItGrad, gradientStep)
            logger.info(Beta)
        if ni > 0:
            free_energyB = 0
            for s in xrange(n_sess):
                free_energyB += vt.Compute_FreeEnergy(y_tilde[s, :, :], m_A[s, :, :], Sigma_A[:, :, :, s], mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C[s, :, :], Sigma_C[:, :, :, s], mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps[s, :], XX[s, :, :, :], W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X[:, s, :, :], Gamma_WX[:, s, :, :], bold=True, S=n_sess)
            if free_energyB < free_energyLA:
                logger.info("free energy has decreased after Beta computation from %f to %f", \
                                free_energyLA, free_energyB)
        if 0 and ni < 5:
            plt.close('all')
            for m in xrange(0, M):
                range_b = np.arange(-10., 20., 0.1)
                beta_plotting = np.zeros_like(range_b)
                grad_plotting = np.zeros_like(range_b)
                for ib, b in enumerate(range_b):
                    beta_plotting[ib] = vt.fun(b, q_Z[m, :, :], Qtilde_sumneighbour[m, :, :],
                                                          neighboursIndexes, gamma)
                    grad_plotting[ib] = vt.grad_fun(b, q_Z[m, :, :], Qtilde_sumneighbour[m, :, :],
                                                     neighboursIndexes, gamma)
                #print beta_plotting
                plt.figure(1)
                plt.hold('on')
                plt.plot(range_b, beta_plotting)
                plt.figure(2)
                plt.hold('on')
                plt.plot(range_b, grad_plotting)
            plt.show()


        # Sigma noise
        if estimateNoise:
            logger.info("M sigma noise step ...")
            for s in xrange(n_sess):
                sigma_eps[s, :] = vt.maximization_sigma_noise_asl(XX[s, :, :, :], m_A[s, :, :], Sigma_A[:, :, :, s], H, m_C[s, :, :], Sigma_C[:, :, :, s], \
                                                    G, Sigma_H, Sigma_G, W, y_tilde[s, :, :], Gamma, \
                                                    Gamma_X[:, s, :, :], Gamma_WX[:, s, :, :], N)

        if PLOT:
            for m in xrange(M):
                SUM_q_Z[m] += [q_Z[m, 1, :].sum()]
                mua1[m] += [mu_Ma[m, 1]]

        free_energy = 0
        for s in xrange(n_sess):
            if s==n_sess-1:
                plotFE = True
            else:
                plotFE = False
            free_energy += vt.Compute_FreeEnergy(y_tilde[s, :, :], m_A[s, :, :], Sigma_A[:, :, :, s], mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C[s, :, :], Sigma_C[:, :, :, s], mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps[s, :], XX[s, :, :, :], W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X[:, s, :, :], Gamma_WX[:, s, :, :],
                                             plot=plotFE, bold=True, S=n_sess)
        if ni > 0:
            if free_energy < free_energyB:
                logger.info("free energy has decreased after Noise computation from %f to %f", free_energyB, free_energy)

        if ni > 0:
            if free_energy < FE[-1]:
                logger.info("WARNING! free energy has decreased in this iteration from %f to %f", FE[-1], free_energy)

        FE += [free_energy]

        if ni > 5:
            #Crit_FE = np.abs((FE[-1] - FE[-2]) / FE[-2])
            FE0 = np.array(FE)
            Crit_FE = np.abs((FE0[-5:] - FE0[-6:-1]) / FE0[-6:-1])
            print Crit_FE
            print (Crit_FE > Thresh * np.ones_like(Crit_FE)).any()
        else:
            Crit_FE = 100

        ni += 1
        cTime += [time.time() - t1]

        logger.info("Computing reconstruction error")
        StimulusInducedSignal = vt.computeFit_asl(H, m_A[s, :, :], G, m_C[s, :, :], W, XX[s, :, :, :])
        rerror = np.append(rerror, \
                           np.mean(((Y[s, :, :] - StimulusInducedSignal) ** 2).sum(axis=0)) \
                           / np.mean((Y[s, :, :] ** 2).sum(axis=0)))

    CompTime = time.time() - t1


    # Normalize if not done already
    if not constraint: # or not normg:
        logger.info("l2-norm of H and G to 1 if not constraint")
        Hnorm = np.linalg.norm(H)
        H /= Hnorm
        Sigma_H /= Hnorm**2
        m_A *= Hnorm
        Gnorm = np.linalg.norm(G)
        G /= Gnorm
        Sigma_G /= Gnorm**2
        m_C *= Gnorm

    if zc:
        H = np.concatenate(([0], H, [0]))

    ## Compute contrast maps and variance
    if computeContrast and len(contrasts) > 0:
        logger.info("Computing contrasts ... ")
        CONTRAST_A, CONTRASTVAR_A, \
        CONTRAST_C, CONTRASTVAR_C = vt.compute_contrasts(condition_names,
                                                         contrasts, m_A[s, :, :], m_C[s, :, :],
                                                         Sigma_A[:, :, :, s], Sigma_C[:, :, :, s], M, J)
    else:
        CONTRAST_A, CONTRASTVAR_A, CONTRAST_C, CONTRASTVAR_C = 0, 0, 0, 0


    ###########################################################################
    ##########################################    PLOTS and SNR computation

    if PLOT:
        logger.info("plotting...")
        print 'FE = ', FE
        vt.plot_convergence(ni, M, cA, cC, cH, cG, cAH, cCG, SUM_q_Z, mua1, muc1, FE)

    logger.info("Nb iterations to reach criterion: %d",  ni)
    logger.info("Computational time = %s min %s s",
                str(np.int(CompTime // 60)), str(np.int(CompTime % 60)))
    logger.info("Iteration time = %s min %s s",
                str(np.int((CompTime // ni) // 60)), str(np.int((CompTime / ni) % 60)))

    logger.info("perfusion baseline mean = %f", np.mean(AL[0, :, s]))
    logger.info("perfusion baseline var = %f", np.var(AL[0, :, s]))
    logger.info("drifts mean = %f", np.mean(AL[1:, :, s]))
    logger.info("drifts var = %f", np.var(AL[1:, :, s]))
    logger.info("noise mean = %f", np.mean(sigma_eps[s, :]))
    logger.info("noise var = %f", np.var(sigma_eps[s, :]))

    SNR10 = 20 * (np.log10(np.linalg.norm(Y[s, :, :]) / \
                np.linalg.norm(Y[s, :, :] - StimulusInducedSignal - PL[s, :, :])))
    logger.info("SNR = %d",  SNR10)

    return ni, m_A.mean(0), H, m_C.mean(0), G, Z_tilde, sigma_eps[s, :], \
           mu_Ma, sigma_Ma, mu_Mc, sigma_Mc, Beta, AL[:, :, s], PL[s, :, :], \
           np.zeros_like(AL[0, :, s]), Sigma_A[:, :, :, s], Sigma_C[:, :, :, s], Sigma_H, Sigma_G, rerror, \
           CONTRAST_A, CONTRASTVAR_A, CONTRAST_C, CONTRASTVAR_C, \
           cA[:], cH[2:], cC[2:], cG[2:], cZ[2:], cAH[2:], cCG[2:], \
           cTime, FE




# -*- coding: utf-8 -*-

import os
import os.path as op
import numpy as np
import glob
import tempfile
import shutil
import logging

import scipy as sp

from nipy.labs.glm import glm
from nipy.modalities.fmri import design_matrix as dm
try:
    from nipy.modalities.fmri.experimental_paradigm import \
        load_paradigm_from_csv_file
except ImportError:
    from nipy.modalities.fmri.experimental_paradigm import \
        load_protocol_from_csv_file as load_paradigm_from_csv_file

import pyhrf
from pyhrf import FmriData
from pyhrf.ndarray import expand_array_in_mask
from pyhrf.tools._io import read_volume, write_volume
from pyhrf.ndarray import xndarray


logger = logging.getLogger(__name__)


def glm_nipy(fmri_data, contrasts=None, hrf_model='Canonical',
             drift_model='Cosine', hfcut=128,
             residuals_model='spherical', fit_method='ols',
             fir_delays=[0],
             rescale_results=False, rescale_factor=None):
    """
    Perform a GLM analysis on fMRI data using the implementation of Nipy.

    Args:
        fmri_data (pyhrf.core.FmriData): the input fMRI data defining the
            paradigm and the measured 3D+time signal.
        contrasts (dict): keys are contrast labels and values are arithmetic
            expressions involving regressor names. Valid names are:
            * names of experimental conditions as defined in fmri_data
            * constant
        hrf_model: "Canonical", "Canonical with Derivative", "FIR"
        residuals_model: "spherical", "ar1"
        fit_method: "ols", "kalman" (If residuals_model is "ar1" then method
            is set to "kalman" and this argument is ignored)
        fir_delays: list of integers indicating the delay of each FIR coefficient
                    (in terms of scans). Eg if TR = 2s. and we want a FIR
                    duration of 20s.: fir_delays=range(10)
    Returns:
        (glm instance, design matrix, dict of contrasts of objects)

    Examples:
    >>> from pyhrf.core import FmriData
    >>> from pyhrf.glm import glm_nipy
    >>> g,dmtx,con = glm_nipy(FmriData.from_vol_ui())
    >>> g,dmtx,con = glm_nipy(FmriData.from_vol_ui(), \
                              contrasts={'A-V':'audio-video'})
    """

    paradigm = fmri_data.paradigm.to_nipy_paradigm()

    # BOLD data
    Y = fmri_data.bold.T
    n_scans = Y.shape[1]

    # Design matrix
    frametimes = np.linspace(0, (n_scans - 1) * fmri_data.tr, n_scans)
    design_matrix = dm.make_dmtx(frametimes, paradigm,
                                 hrf_model=hrf_model,
                                 drift_model=drift_model, hfcut=hfcut,
                                 fir_delays=fir_delays)

    ns, nr = design_matrix.matrix.shape
    logger.info('Design matrix built with %d regressors:', nr)
    for rn in design_matrix.names:
        logger.info('    - %s', rn)

    # GLM fit
    my_glm = glm.glm()
    logger.info('Fit GLM - method: %s, residual model: %s', fit_method,
                residuals_model)
    my_glm.fit(Y.T, design_matrix.matrix, method=fit_method,
               model=residuals_model)

    from pyhrf.tools import map_dict
    from pyhrf.paradigm import contrasts_to_spm_vec

    if rescale_results:
        if 1:
            if rescale_results and rescale_factor is None:
                # Rescale by the norm of each regressor in the design matrix
                dm_reg_norms = (design_matrix.matrix ** 2).sum(0) ** .5
                logger.info('GLM results (beta and con effects) are '
                            'rescaled by reg norm. Weights: %s ',
                            str(dm_reg_norms))

                for ib in xrange(my_glm.beta.shape[0]):
                    my_glm.beta[ib] = my_glm.beta[ib] * dm_reg_norms[ib]

            else:
                logger.info('GLM results (beta and con effects) are '
                            'rescaled by input scale factor.')

                # Use input rescale factors:
                for ib in xrange(rescale_factor.shape[0]):
                    my_glm.beta[ib] = my_glm.beta[ib] * rescale_factor[ib]
                    # TOCHECK: nvbeta seems to be a covar matrix between reg
                    # -> we dont get position-specific variances ...
                    #my_glm.nvbeta[ib,:] = my_glm.nvbeta[ib,:] * rescale_factor[ib]**2

    if contrasts is not None:
        con_vectors = contrasts_to_spm_vec(design_matrix.names, contrasts)
        # if rescale_results:
        #     for con_vec in con_vectors.itervalues():
        #         con_vec *= dm_reg_norms
        contrast_result = map_dict(my_glm.contrast, con_vectors)
    else:
        contrast_result = None

    return my_glm, design_matrix, contrast_result

# actually: not possible to compute PPM from glm results
# Should relaunch estimation with propoer model under SPM
# def PPMcalculus_glmWN(beta, var_beta, dm, threshold_value):
    '''
    #Function to compute PPM after glm estimation with white noise
    #Inputs:
    #-beta: beta values
    #-var_beta: normalized variance of beta
    #-dm: design matrix
    #-threshol value: phyisiologically relevant size of effect
            #--> take a percentage of mean signal value
    #GLM estimation must be performed with nipy
    #Compute the pp for one voxel!
    #'''
    #from scipy.integrate import quad

    #X = dm.matrix
    #mean_apost = 1./(np.dot(dm.transpose(), dm)*1./(var_beta) + 1./beta)
    #var_apost = mean_apost*np.dot(dm.transpose(), dm)*1./(var_beta)

    #f_distrib = lambda t: 1/np.sqrt(2*np.pi*var_apost**2)*np.exp(- (t - mean_apost)**2 / (2*var_apost**2) )
    #Proba = quad(f_distrib, threshold_value, float('inf'))[0]

    # return Proba


def glm_nipy_from_files(bold_file, tr, paradigm_csv_file, output_dir,
                        mask_file, session=0, contrasts=None,
                        con_test_baseline=0.0,
                        hrf_model='Canonical',
                        drift_model='Cosine', hfcut=128,
                        residuals_model='spherical', fit_method='ols',
                        fir_delays=[0]):
    """
    #TODO: handle surface data
    hrf_model : Canonical | Canonical with Derivative | FIR

    """

    fdata = FmriData.from_vol_files(mask_file, paradigm_csv_file,
                                    [bold_file], tr)
    g, dm, cons = glm_nipy(fdata, contrasts=contrasts, hrf_model=hrf_model,
                           hfcut=hfcut, drift_model=drift_model,
                           residuals_model=residuals_model,
                           fit_method=fit_method, fir_delays=fir_delays)

    ns, nr = dm.matrix.shape
    cdesign_matrix = xndarray(dm.matrix, axes_names=['time', 'regressor'],
                              axes_domains={'time': np.arange(ns) * tr,
                                            'regressor': dm.names})

    cdesign_matrix.save(op.join(output_dir, 'design_matrix.nii'))

    beta_files = []
    beta_values = dict.fromkeys(dm.names)
    beta_vars = dict.fromkeys(dm.names)
    beta_vars_voxels = dict.fromkeys(dm.names)
    for ib, bname in enumerate(dm.names):
        # beta values
        beta_vol = expand_array_in_mask(g.beta[ib], fdata.roiMask > 0)
        beta_fn = op.join(output_dir, 'beta_%s.nii' % bname)
        write_volume(beta_vol, beta_fn, fdata.meta_obj)
        beta_files.append(beta_fn)
        beta_values[bname] = beta_vol

        # normalized variance of betas
        # variance: diag of cov matrix
        beta_vars[bname] = sp.diag(g.nvbeta)[ib]
        # sig2 = g.s2 #ResMS
        # variance for all voxels, condition ib
        var_cond = sp.diag(g.nvbeta)[ib] * g.s2
        beta_vars_voxels[bname] = var_cond
        #beta_var_fn = op.join(output_dir, 'var_beta_%s.nii' %bname)
        #write_volume(beta_var, beta_var_fn, fdata.meta_obj)
        # beta_var_files.append(beta_var_fn)

    if cons is not None:
        con_files = []
        pval_files = []
        for cname, con in cons.iteritems():
            con_vol = expand_array_in_mask(con.effect, fdata.roiMask > 0)
            con_fn = op.join(output_dir, 'con_effect_%s.nii' % cname)
            write_volume(con_vol, con_fn, fdata.meta_obj)
            con_files.append(con_fn)

            pval_vol = expand_array_in_mask(con.pvalue(con_test_baseline),
                                            fdata.roiMask > 0)
            pval_fn = op.join(output_dir, 'con_pvalue_%s.nii' % cname)
            write_volume(pval_vol, pval_fn, fdata.meta_obj)
            pval_files.append(pval_fn)
    else:
        con_files = None
        pval_files = None

    dof = g.dof
    # if do_ppm:
    # for

    # TODO: FIR stuffs
    # , con_files, pval_files
    return beta_files, beta_values, beta_vars_voxels, dof

    # Paradigm

    # Fix bug in nipy/modalities/fmri/experimental_paradigm.py
    # -> amplitudes are not converted to floats
    # @@ -187,7 +187,7 @@ def load_paradigm_from_csv_file(path, session=None):
    #          if len(row) > 3:
    #              duration.append(float(row[3]))
    #          if len(row) > 4:
    # -            amplitude.append(row[4])
    # +            amplitude.append(float(row[4]))


if pyhrf.cfg['global']['spm_path'] is not None:
    def glm_matlab_from_files(bold_file, tr, paradigm_csv_file, output_dir,
                              mask_file, hf_cut=128, hack_mask=False):
        """
        Only mono-session
        #TODO: compute mask if mask_file does not exist
        #TODO: handle contrasts
        """

        # Functional mask
        # if not op.exists(mask_file):
        #     logger.info('Mask file does not exist. Computing mask from '
        #                 'BOLD data')
        #     compute_mask_files(bold_files, mask_file, False, 0.4, 0.9)

        # split BOLD into 3D vols:
        bold, hbold = read_volume(bold_file)
        bold_files = []
        tmp_path = tempfile.mkdtemp(dir=pyhrf.cfg['global']['tmp_path'])
        for iscan, bscan in enumerate(bold):
            f = op.join(tmp_path, 'bold_%06d.nii' % iscan)
            write_volume(bscan, f, hbold)
            bold_files.append(f)
        bold_files = ';'.join(bold_files)

        script_path = op.join(op.dirname(pyhrf.__file__), '../../script/SPM')
        spm_path = pyhrf.cfg['global']['spm_path']
        matlab_code = "cd %s;paradigm_file='%s';TR=%f;mask_file='%s';" \
            "bold_files='%s';output_path='%s';" \
            "HF_cut=%f;spm_path='%s';api=1;hack_mask=%d;glm_intra_subj;exit" \
            % (script_path, paradigm_csv_file, tr, mask_file, bold_files, output_dir,
               hf_cut, spm_path, hack_mask)

        matlab_cmd = 'matlab -nosplash -nodesktop -r "%s"' % matlab_code

        if op.exists(op.join(output_dir, 'SPM.mat')):
            # remove SPM.mat so that SPM won't ask over ask overwriting
            os.remove(op.join(output_dir, 'SPM.mat'))

        # print 'matlab cmd:'
        # print matlab_cmd
        os.system(matlab_cmd)

        # Fix shape of outputs if necessary
        # eg if input data has shape (n,m,1) then SPM will write outputs of
        # shape (n,m) so that they are not consistent with their QForm
        input_shape = bscan.shape
        for foutput in glob.glob(op.join(output_dir, '*.img')):
            data, h = read_volume(foutput)
            if data.ndim < 3:
                sm = ','.join([[':', 'np.newaxis'][d == 1]
                               for d in input_shape])
                exec('data = data[%s]' % sm)
                assert data.shape == input_shape
                write_volume(data, foutput, h)

        shutil.rmtree(tmp_path)
        # TODO: maybe find a better way to grab beta file names
        beta_files = sorted(glob.glob(op.join(output_dir, 'beta_*.img')))
        return beta_files



# -*- coding: utf-8 -*-

import unittest
import tempfile
import shutil
import logging

import numpy as np

import pyhrf
import pyhrf.vbjde.vem_tools as vt

from pyhrf import FmriData
from pyhrf.jde.models import simulate_bold
from pyhrf.boldsynth.hrf import getCanoHRF
try:
    from collections import OrderedDict
except ImportError:
    from pyhrf.tools.backports import OrderedDict


logger = logging.getLogger(__name__)


class VEMToolsTest(unittest.TestCase):

    def setUp(self):
        np.random.seed(8652761)

        tmpDir = tempfile.mkdtemp(prefix='pyhrf_tests',
                                  dir=pyhrf.cfg['global']['tmp_path'])
        self.tmp_dir = tmpDir
        self.clean_tmp = True
        simu = simulate_bold(self.tmp_dir, spatial_size='random_small')
        self.data_simu = FmriData.from_simulation_dict(simu)

    def tearDown(self):
        if self.clean_tmp:
            logger.info('Remove tmp dir %s', self.tmp_dir)
            shutil.rmtree(self.tmp_dir)
        else:
            logger.info('Keep tmp dir %s' % self.tmp_dir)

    def test_computeFit(self):
        X = OrderedDict([])
        M = 51
        N = 325
        J = 25
        Thrf = 25.
        dt = .5
        TT, m_h = getCanoHRF(Thrf, dt)
        m_A = np.zeros((J, M), dtype=np.float64)
        m_H = np.array(m_h).astype(np.float64)
        stimIndSignal = vt.computeFit(m_H, m_A, X, J, N)

    def test_entropyA(self):
        M = 51
        J = 25
        Sigma_A = np.zeros((M, M, J), np.float64)
        entropy = vt.A_Entropy(Sigma_A, M, J)

    def test_entropyH(self):
        D = 3
        Sigma_H = np.ones((D, D), dtype=np.float64)
        entropy = vt.H_Entropy(Sigma_H, D)

    def test_entropyZ(self):
        M = 51
        K = 2
        J = 25
        q_Z = np.zeros((M, K, J), dtype=np.float64)
        entropy = vt.Z_Entropy(q_Z, M, J)

    def test_max_sigmaH(self):
        D = 3
        Thrf = 25.
        dt = .5
        TT, m_h = getCanoHRF(Thrf, dt)
        m_h = m_h[:D]
        m_H = np.array(m_h).astype(np.float64)
        Sigma_H = np.ones((D, D), dtype=np.float64)
        order = 2
        D2 = vt.buildFiniteDiffMatrix(order, D)
        R = np.dot(D2, D2) / pow(dt, 2*order)
        sigmaH = vt.maximization_sigmaH(D, Sigma_H, R, m_H)

    def test_max_sigmaH_prior(self):
        D = 3
        Thrf = 25.
        dt = .5
        TT, m_h = getCanoHRF(Thrf, dt)
        m_h = m_h[:D]
        m_H = np.array(m_h).astype(np.float64)
        Sigma_H = np.ones((D, D), dtype=np.float64)
        order = 2
        D2 = vt.buildFiniteDiffMatrix(order, D)
        R = np.dot(D2, D2) / pow(dt, 2*order)
        gamma_h = 1000
        sigmaH = vt.maximization_sigmaH_prior(D, Sigma_H, R, m_H, gamma_h)

    def test_matrix(self):
        Thrf = 25.
        dt = .5
        TT, m_h = getCanoHRF(Thrf, dt)
        m_H = np.array(m_h)
        m = vt.mult(m_H, m_H)

    def test_maximum(self):
        M = 51
        K = 2
        vector = np.ones((K, 1), dtype=np.float64)
        vector[0] = 2
        maxs, maxs_ind = vt.maximum(vector)

    def test_normpdf(self):
        y = vt.normpdf(1, 1, 1)

    def test_polyFit(self):
        N = 325
        TR = 1.
        data = self.data_simu
        Y = data.bold
        P = vt.PolyMat(N, 4, TR)
        L = vt.polyFit(Y, TR, 4, P)

    def test_PolyMat(self):
        N = 325
        TR = 1.
        P = vt.PolyMat(N, 4, TR)

    def test_compute_mat_X2(self):
        N = 325
        TR = 1.
        D = 3
        dt = .5
        data = self.data_simu
        X = OrderedDict([])
        Onsets = data.get_joined_onsets()
        for condition, Ons in Onsets.iteritems():
            X[condition] = vt.compute_mat_X_2(N, TR, D, dt, Ons)

    def test_buildFiniteDiffMatrix(self):
        order = 2
        D = 3
        diffMat = vt.buildFiniteDiffMatrix(order, D)



# -*- coding: utf-8 -*-

import logging

import numpy as np

import pyhrf

from pyhrf.ndarray import xndarray
from pyhrf.tools import array_summary


logger = logging.getLogger(__name__)


class Trajectory:
    """ Keep track of a numpy array that is modified _inplace_ iteratively
    TODO: when mature, should be moved to pyhrf.ndarray should replace pyhrf.jde.samplerbase.Trajectory
    """

    def __init__(self, variable, history_pace, history_start, max_iterations,
                 init_iteration=None, axes_names=None, axes_domains=None):
        """
        Args:
            *variable* is the tracked numpy array that has to be
            modified _inplace_ else the reference to it is lost.
            *history_pace* is the pace at which each value of *variable* is saved.
            *history_start* is the iteration when to start saving values of
                            *variable*
            *max_iterations* is the total maximal number of iterations
                           (for allocating memory)
            *init_iteration* sets the integer value of the first
                             saved iteration. -1 is useful to mark the init
                             value which is prior to the Oth iteration.
                             If None then the value of *variable* at the creation
                             of the Trajectory is no saved.
            *axes_names* is a list of string for the axes of *variable*.
                         Only used in *to_cuboid*
            *axes_domains* is a dict of numpy arrays for the domains of the axes.
                         Only used in *to_cuboid*
        """
        self.variable = variable
        self.axes_names = axes_names
        self.axes_domains = axes_domains
        self.hist_pace = history_pace
        self.hist_start = history_start

        if history_pace > 0:
            nsamples_max = (max_iterations - history_start) / history_pace + \
                           (((max_iterations - history_start) %
                             history_pace) > 0)
        else:
            nsamples_max = 0

        if init_iteration is not None:
            nsamples_max += 1  # +1 because of init

        self.history = np.zeros((nsamples_max,) + variable.shape,
                                dtype=variable.dtype)

        self.sample_count = 0
        self.saved_iterations = []
        if init_iteration is not None:
            #-1 is initialisation.
            self.history[0] = self.variable[:]
            self.saved_iterations.append(init_iteration)
            self.sample_count = 1

    def update(self, iteration):
        """ Record the current variable value """
        if self.hist_pace > 0 and iteration >= self.hist_start and \
                ((iteration - self.hist_start) % self.hist_pace) == 0:
            logger.debug('-> record: %s (sample_count: %d, hist shape: %s)',
                         str(self.variable), self.sample_count,
                         str(self.history.shape))
            self.history[self.sample_count] = self.variable[:]
            self.saved_iterations.append(iteration)
            self.sample_count += 1
        else:
            logger.debug("-> don't record (self.hist_pace: %d, "
                         "self.hist_start: %d)", self.hist_pace, self.hist_start)

    def get_last(self):
        """ Return the last saved element """
        if self.sample_count > 0:
            return self.history[self.sample_count - 1]
        else:
            raise Exception('No recorded value')

    def to_cuboid(self):
        """ Pack the current trajectory in a xndarray
        """
        if self.sample_count > 0:
            if self.axes_names is not None:
                an = ['iteration'] + self.axes_names
            else:
                an = ['iteration'] + \
                    ['axis_%02d' % i for i in xrange(self.variable.ndim)]

            if self.axes_domains is not None:
                ad = self.axes_domains.copy()
            else:
                ad = {}
            ad['iteration'] = np.array(self.saved_iterations, dtype=int)
            c = xndarray(self.history, axes_names=an, axes_domains=ad)
            return c
        else:
            return None  # raise Exception ?


### Gibbs Sampler ###
# When mature, the following should replace pyhrf.jde.samplerbase
# and be moved to pyhrf.stats.gibbs


class GSVariable:

    def __init__(self, name, initialization, do_sampling=True,
                 axes_names=None, axes_domains=None):
        """
        Create a variable to be used in a Gibbs Sampler.

        Args:
            *name* (str): the label of the variable. Must not be the same
                as another variable in the model.
            *initialization* (str or numpy.ndarray):
                If numpy.ndarray then this value will be used as init value
                for the sampling.
                If str then available choices are:
                    'random': init will be defined by method self.get_random_init
                    'custom': init will be defined by method self.get_custom_init
                    'truth': init will be set to true value
                             (see method self.set_true_value)
            *do_sampling* (bool): flag to enable sampling.
                If False then the variable will be constant and set to its
                init value.
            *axes_names* (list): names for the axes. Used for outputs
            *axes_domains* (dict): domains associated to axes. Used for outputs
        """
        self.name = name
        self.check_initialization_arg(initialization)
        self.initialization = initialization
        self.do_sampling = do_sampling
        self.axes_names = axes_names
        self.axes_domains = axes_domains
        if self.axes_domains is None:
            self.axes_domains = {}

        self.true_value = None

    def check_initialization_arg(self, ia):
        if not (isinstance(ia, (str, np.ndarray)) or
                np.isscalar(ia)):
            raise Exception('Wrong type %s for initialization argument.'
                            'Allowed: str, npumpy.ndarray, scalar' % ia.__class__)

        if isinstance(ia, str):
            choices = ['random', 'custom', 'truth']
            if not ia in choices:
                raise Exception('Wrong choice for initialization argument: %s. '
                                'Allowed: %s' % (ia, ','.join(choices)))

    def enable_sampling(self, flag=True):
        self.do_sampling = flag

    def set_initialization(self, init):
        self.check_initialization_arg(init)
        self.initialization = init

    def set_true_value(self, true_value):
        if self.axes_names is not None:
            assert self.true_value.ndim == len(self.axes_names)

        self.true_value = true_value

    def get_variable(self, vname):
        return self.sampler.get_variable(vname)

    def get_variable_value(self, vname):
        """
        Short-hand to get variable among all those defined
        in the parent sampler
        """
        return self.sampler.get_variable_value(vname)

    # Initialization methods

    def set_init_value(self):
        """ Set the initial value of self.current_value, depending on the
        *initialization* scenario (random, custom, truth).
        """

        if isinstance(self.initialization, str):
            if self.initialization == 'random':
                self.current_value = self.get_random_init()
            elif self.initialization == 'custom':
                self.current_value = self.get_custom_init()
            elif self.initialization == 'truth':
                if self.true_value is None:
                    raise Exception('Missing true value to init variable %s'
                                    % (self.name))
                self.current_value = self.true_value
        elif isinstance(self.initialization, np.ndarray) or \
                np.isscalar(self.initialization):
            self.current_value = self.initialization

        if np.isscalar(self.current_value):
            self.current_value = np.array([self.current_value])

        if not isinstance(self.current_value, np.ndarray):
            raise Exception('Initialization of variable "%s" '
                            'with scenario "%s" did not returned a '
                            'a numpy.ndarray, but: %s'
                            % (self.name, self.initialization,
                               str(self.current_value)))

        if self.current_value.ndim == 0:  # don't handle 0-d arrays
            self.current_value = self.current_value[np.newaxis]

        if self.axes_names is not None:
            assert self.current_value.ndim == len(self.axes_names)

    def get_random_init(self):
        """
        Must return a random numpy.ndarray that will then be used as
        init value for sampling. For example, it can be a sample from
        the prior distribution.
        This function will also be used to test for the sensitivity to
        initialization.
        """
        raise NotImplementedError()

    def get_custom_init(self):
        """
        Must return a numpy.ndarray. Consider initializing with a good
        guess so that sampling converges more quickly.
        """
        raise NotImplementedError()

    def _init_sampling(self, sampler):

        self.sampler = sampler
        self.set_init_value()

        if self.do_sampling:
            # init default observables (mean, variance):
            self.cumul = np.zeros(self.current_value.shape)
            self.cumul_disp = np.zeros(self.current_value.shape, dtype=float)
            self.obs_mean = np.zeros(self.current_value.shape, dtype=float)
            self.obs_var = np.zeros(self.current_value.shape, dtype=float)
            self.nb_its_obs = 0.  # to compute obs mean and var during sampling

            self.init_observables()

            # Enable tracking of samples, mean and variance:
            self.track_sampled_quantity(self.name + '_hist_smpl',
                                        self.current_value)
            self.track_obs_quantity(
                self.name + '_hist_obs_mean', self.obs_mean)
            self.track_obs_quantity(self.name + '_hist_obs_var', self.obs_var)

            self.init_sampling()

    def track_sampled_quantity(self, name, quantity, history_pace=None,
                               axes_names=None, axes_domains=None):

        self.sampler.track_sampled_quantity(name, quantity, history_pace,
                                            axes_names, axes_domains)

    def track_obs_quantity(self, name, quantity, history_pace=None,
                           axes_names=None, axes_domains=None):

        self.sampler.track_obs_quantity(name, quantity, history_pace,
                                        axes_names, axes_domains)

    def reset(self):
        pass

    def init_observables(self):
        pass

    def init_sampling(self):
        pass

    # Sampling methods
    def _sample_next(self):
        """ Internal method called by parent GibbsSampler to perform a sampling
        iteration
        """

        if self.do_sampling:
            v = self.sample()
            if np.isscalar(v):
                v = np.array([v])
            if v.shape != self.current_value.shape:
                raise Exception('Method *sample* returned an array of shape '
                                'inconsistent with previous value: got %s,'
                                'expected %s' % (str(v.shape),
                                                 str(self.current_value.shape)))
            # inplace modification so that current_value can be trackable
            # by Trajectory object
            self.current_value[:] = v
            logger.debug(
                'sample of %s: %s', self.name, str(self.current_value))

    def sample(self):
        """
        Draw a sample conditionally to the current Gibbs Sampler state.
        Must return a numpy.ndarray.

        Variables which have been registered in the parent GibbsSampler object
        can be retrieved via methods self.get_variable(var_name) and
        self.get_variable_value(var_name)
        """
        raise NotImplementedError()

    def _update_observables(self):
        self.nb_its_obs += 1.
        self.cumul += self.current_value
        np.divide(self.cumul, self.nb_its_obs, self.obs_mean)
        self.cumul_disp += (self.current_value - self.obs_mean) ** 2
        np.divide(self.cumul_disp, self.nb_its_obs, self.obs_var)

        self.update_observables()

    def update_observables(self):
        """ Update quantities after the burnin period """
        pass

    # End of sampling

    def _finalize_sampling(self):
        pass

    def get_estim_value_for_check(self):
        return self.obs_mean

    def get_true_value_for_check(self):
        return self.true_value

    def get_accuracy_against_truth(self, abs_error, rel_error, fv, tv,
                                   atol, rtol):
        """ Return the accuray of the estimate *fv*, compared to the true
        value *tv*
        """
        # same criterion as np.allclose:
        acc = abs_error <= (atol + rtol * np.maximum(np.abs(tv),
                                                     np.abs(fv)))
        return self.axes_names, acc

    def check_against_truth(self, atol, rtol, inaccuracy_handling='print'):

        fv = self.get_estim_value_for_check()
        tv = self.get_true_value_for_check()

        if tv is None:
            logger.info(
                'Warning: no true val to check against for %s', self.name)
        elif self.do_sampling:  # when sampling is off, assume there is no need
                               # to check against truth

            abs_error = np.abs(tv - fv)
            rel_error = abs_error / np.maximum(np.abs(tv), np.abs(fv))

            acc = self.get_accuracy_against_truth(abs_error, rel_error, fv, tv,
                                                  atol, rtol)
            is_accurate = acc[1].all()
            logger.info('Check error for "%s" -> '
                        'estim: %s, truth: %s, atol=%f, rtol=%f', self.name,
                        str(fv), str(tv), atol, rtol)

            logger.info('Fit error for "%s": avg aerr=%f, avg rerr=%f, '
                        'is_accurate=%s', self.name, abs_error.mean(),
                        rel_error.mean(), is_accurate)
            if not is_accurate:
                m = "Final value of %s is not close to " \
                    "true value.\n -> aerror: %s\n -> rerror: %s\n" \
                    " Final value:\n %s\n True value:\n %s\n" \
                    % (self.name, array_summary(abs_error),
                       array_summary(rel_error), str(fv), str(tv))
                if inaccuracy_handling == 'raise':
                    raise Exception(m)
                elif inaccuracy_handling == 'print':
                    print '\n'.join(['!! ' + s for s in m.split('\n')])

            self.truth_checking_report = {
                'abs_error': abs_error,
                'rel_error': rel_error}

    def _get_outputs(self, output_type='ndarray'):
        """
        output_type : 'ndarray' or 'cuboid'
        """

        outputs = {}

        # Default outputs: post. mean, post. var, trajectories.
        on = self.name + '_obs_mean'
        if output_type == 'cuboid':
            outputs[on] = xndarray(self.obs_mean, axes_names=self.axes_names,
                                   axes_domains=self.axes_domains)
        else:
            outputs[on] = self.obs_mean

        on = self.name + '_obs_var'
        if output_type == 'cuboid':
            outputs[on] = xndarray(self.obs_var, axes_names=self.axes_names,
                                   axes_domains=self.axes_domains)
        else:
            outputs[on] = self.obs_var

        # Custom outputs:
        self.set_outputs(outputs, output_type)

        return outputs

    def set_outputs(self, outputs, output_type='ndarray'):
        """
        Args:
            - outputs (dict): dictionary to be updated with custom outputs.
            - output_type (str): 'ndarray' or 'cuboid'
        Return: None
        """
        pass


class GibbsSampler:

    def __init__(self, sampled_variables, nb_its_max, obs_pace=1, burnin=.3,
                 sample_hist_pace=-1, obs_hist_pace=-1,):

        self.variables = {}
        self.sampled_variables = sampled_variables

        for v in sampled_variables:
            self.set_variable(v.name, v)

        def get_fraction_or_nb(nb, tot):
            if nb > 0. and nb < 1.:
                return int(round(tot * nb))
            else:
                return nb

        self.nb_its_max = nb_its_max
        self.burnin = get_fraction_or_nb(burnin, nb_its_max)
        self.smpl_hist_pace = get_fraction_or_nb(sample_hist_pace, nb_its_max)
        self.obs_hist_pace = get_fraction_or_nb(obs_hist_pace, nb_its_max)
        self.tracked_quantities = {}

        logger.info('GibbsSampler init. Burnin: %d, nb_its_max: %d, '
                    'smpl_hist_pace: %d, obs_hist_pace: %d,', self.burnin,
                    self.nb_its_max, self.smpl_hist_pace, self.obs_hist_pace)

        # TODO: global hist pace

    def set_variable(self, name, var):
        if not (isinstance(var, (np.ndarray, GSVariable)) or
                np.isscalar(var)):
            raise Exception('Wrong variable type: %s. Allowed: numpy.ndarray, '
                            'GSVariable or scalar' % var.__class__)

        if self.variables.has_key(name):
            raise Exception('Variable %s already registered')

        self.variables[name] = var

    def set_variables(self, var_dict):
        for k, v in var_dict.iteritems():
            self.set_variable(k, v)

    def get_variable(self, vname):
        v = self.variables.get(vname, None)
        if v is None:
            raise Exception('Unregistered variable %s' % vname)
        else:
            return v

    def get_variable_value(self, vname):

        v = self.get_variable(vname)
        if isinstance(v, np.ndarray) or np.isscalar(v):
            return v
        elif isinstance(v, GSVariable):
            return v.current_value

    def set_true_values(self, true_values):
        for vname, true_value in true_values.iteritems():
            self.get_variable(vname).set_true_value(true_value)

    def set_true_value(self, vname, true_value):
        self.get_variable(vname).set_true_value(true_value)

    def set_initialization(self, vname, init):
        self.get_variable(vname).set_initialization(init)

    def reset(self):
        """
        Reset the Gibbs Sampler:
            - remove all previous history of quantities (trajectories)
            - call reset method of all variables
        """
        self.tracked_quantities = {}
        for v in self.sampled_variables:
            v.reset()

    def stop_criterion(self, iteration):
        return False

    def iterate_sampling(self):
        it = 0
        while it < self.nb_its_max and not self.stop_criterion(it):
            yield it
            it += 1

    def _track_quantity(self, q, name, axes_names, axes_domains,
                        history_pace, hist_start):
        if not self.tracked_quantities.has_key(name):
            trajectory = Trajectory(q, history_pace, hist_start, self.nb_its_max,
                                    axes_names, axes_domains)
            self.tracked_quantities[name] = trajectory
        else:
            raise Exception('Quantity %s already tracked' % name)

    def track_sampled_quantity(self, name, q, history_pace=None, axes_names=None,
                               axes_domains=None):
        if history_pace is None:
            history_pace = self.smpl_hist_pace

        self._track_quantity(q, name, axes_names, axes_domains, history_pace,
                             hist_start=0)

    def track_obs_quantity(self, name, q, history_pace=None, axes_names=None,
                           axes_domains=None):

        if history_pace is None:
            history_pace = self.obs_hist_pace

        self._track_quantity(q, name, axes_domains, axes_domains,
                             history_pace, hist_start=self.burnin)

    def run(self):

        for v in self.sampled_variables:
            v._init_sampling(self)

        for it in self.iterate_sampling():

            for v in self.sampled_variables:
                v._sample_next()

                if it >= self.burnin:
                    v._update_observables()

            for tname, trajectory in self.tracked_quantities.iteritems():
                logger.debug('Update Trajectory "%s" (it=%d) ...', tname, it)
                trajectory.update(it)

            #self.callback(it, self)

        for v in self.sampled_variables:
            v._finalize_sampling()

    def check_against_truth(self, default_atol=0.1, default_rtol=0.1,
                            var_specific_atol=None, var_specific_rtol=None,
                            inaccuracy_handling='print'):
        """
        """
        if var_specific_rtol is None:
            var_specific_rtol = {}
        if var_specific_atol is None:
            var_specific_atol = {}

        for v in self.sampled_variables:
            atol = var_specific_atol.get(v.name, default_atol)
            rtol = var_specific_rtol.get(v.name, default_rtol)
            v.check_against_truth(atol=atol, rtol=rtol,
                                  inaccuracy_handling=inaccuracy_handling)

    def get_outputs(self, output_type='ndarray'):
        """
        output_type : 'ndarray' or 'cuboid'
        """

        outputs = {}

        for v in self.sampled_variables:
            self._update_outputs(outputs, v._get_outputs(output_type))

        for qname, q in self.tracked_quantities.iteritems():
            if output_type == 'cuboid':
                ctrajectory = q.to_cuboid()
                if ctrajectory is not None:
                    self._add_output(outputs, qname, ctrajectory)
                else:
                    logger.info("Trajectory '%s' -> no output")
            else:
                self._add_output(outputs, qname, q.history)

        return outputs

    def _add_output(self, cur_outputs, new_output_name, new_output_value):
        if cur_outputs.has_key(new_output_name):
            raise Exception('Output %s already defined' % new_output_name)
        else:
            cur_outputs[new_output_name] = new_output_value

    def _update_outputs(self, cur_outputs, new_outputs):
        for k, v in new_outputs.iteritems():
            self._add_output(cur_outputs, k, v)



# -*- coding: utf-8 -*-

import os.path as op
import shutil
import unittest
import logging

import numpy as np
import numpy.testing as npt

from scipy.stats import truncnorm

import pyhrf
import pyhrf.sandbox.physio as phy

from pyhrf import Condition
from pyhrf.tools import is_importable


logger = logging.getLogger(__name__)


class SimulationTest(unittest.TestCase):

    def setUp(self):
        # called before any unit test of the class
        self.tmp_path = pyhrf.get_tmp_path()  # create a temporary folder
        self.clean_tmp = True

    def tearDown(self):
        # called after any unit test of the class
        if self.clean_tmp:
            logger.info('cleaning temporary folder ...')
            shutil.rmtree(self.tmp_path)

    @unittest.skipUnless(is_importable("PIL"), "Pillow (optional dep) is N/A")
    def test_simulate_asl_full_physio(self):

        r = phy.simulate_asl_full_physio()
        # let's just test the shapes of objects and the presence of some
        # physio-specific simulation items
        item_names = r.keys()
        self.assertIn('perf_stim_induced', item_names)
        self.assertIn('flow_induction', item_names)
        self.assertIn('perf_stim_induced', item_names)
        self.assertEqual(r['labels_vol'].shape, (3, 1, 2, 2))

    @unittest.skipUnless(is_importable("PIL"), "Pillow (optional dep) is N/A")
    def test_simulate_asl_full_physio_outputs(self):

        phy.simulate_asl_full_physio(self.tmp_path)

        def makefn(fn):
            return op.join(self.tmp_path, fn)

        self.assertTrue(op.exists(makefn('flow_induction.nii')))
        self.assertTrue(op.exists(makefn('neural_efficacies_audio.nii')))

    @unittest.skipUnless(is_importable("PIL"), "Pillow (optional dep) is N/A")
    def test_simulate_asl_physio_rfs(self):

        r = phy.simulate_asl_physio_rfs()
        # let's just test the shapes of objects and the presence of some
        # physio-specific simulation items:
        item_names = r.keys()
        self.assertIn('perf_stim_induced', item_names)
        self.assertIn('primary_brf', item_names)
        self.assertIn('perf_stim_induced', item_names)
        self.assertEqual(r['labels_vol'].shape, (3, 1, 2, 2))
        self.assertEqual(r['bold'].shape[1], 4)  # flat spatial axis

    def test_create_tbg_neural_efficacies(self):
        """ Test the generation of neural efficacies from a truncated
        bi-Gaussian mixture
        """
        np.random.seed(25432)
        m_act = 5.
        v_act = .05
        v_inact = .05
        cdef = [Condition(m_act=m_act, v_act=v_act, v_inact=v_inact)]
        npos = 5000
        labels = np.zeros((1, npos), dtype=int)
        labels[0, :npos / 2] = 1
        phy_params = phy.PHY_PARAMS_FRISTON00
        ne = phy.create_tbg_neural_efficacies(phy_params, cdef, labels)

        # check shape consistency:
        self.assertEqual(ne.shape, labels.shape)

        # check that moments are close to theoretical ones
        ne_act = ne[0, np.where(labels[0])]
        ne_inact = ne[0, np.where(labels[0] == 0)]
        m_act_theo = truncnorm.mean(0, phy_params['eps_max'], loc=m_act,
                                    scale=v_act ** .5)
        v_act_theo = truncnorm.var(0, phy_params['eps_max'], loc=m_act,
                                   scale=v_act ** .5)
        (ne_act.mean(), m_act_theo)
        npt.assert_approx_equal(ne_act.var(), v_act_theo, significant=2)

        m_inact_theo = truncnorm.mean(0, phy_params['eps_max'], loc=0.,
                                      scale=v_inact ** .5)
        v_inact_theo = truncnorm.var(0, phy_params['eps_max'], loc=0.,
                                     scale=v_inact ** .5)
        npt.assert_approx_equal(ne_inact.mean(), m_inact_theo, significant=2)
        npt.assert_approx_equal(ne_inact.var(), v_inact_theo, significant=2)
        npt.assert_array_less(ne, phy_params)
        npt.assert_array_less(0., ne)

    def test_create_physio_brf(self):
        phy_params = phy.PHY_PARAMS_FRISTON00
        dt = .5
        duration = 25.
        brf = phy.create_physio_brf(phy_params, response_dt=dt,
                                    response_duration=duration)

        if 0:
            import matplotlib.pyplot as plt
            t = np.arange(brf.size) * dt
            plt.plot(t, brf)
            plt.title('BRF')
            plt.show()

        npt.assert_approx_equal(brf[0], 0., significant=4)
        npt.assert_array_almost_equal(brf[-1], 0., decimal=4)

        npt.assert_approx_equal(np.argmax(brf) * dt, 3.5, significant=5)

    def test_create_physio_prf(self):

        phy_params = phy.PHY_PARAMS_FRISTON00
        dt = .5
        duration = 25.
        prf = phy.create_physio_prf(phy_params, response_dt=dt,
                                    response_duration=duration)

        if 0:
            import matplotlib.pyplot as plt
            t = np.arange(prf.size) * dt
            plt.plot(t, prf)
            plt.title('PRF')
            plt.show()

        npt.assert_approx_equal(prf[0], 0., significant=4)
        npt.assert_array_almost_equal(prf[-1], 0., decimal=4)

        npt.assert_approx_equal(np.argmax(prf) * dt, 2.5, significant=5)

    def test_create_evoked_physio_signal(self):
        import pyhrf.paradigm

        phy_params = phy.PHY_PARAMS_FRISTON00
        tr = 1.
        duration = 20.
        ne = np.array([[10., 5.]])
        nb_conds, nb_vox = ne.shape
        # one single stimulation at the begining
        paradigm = pyhrf.paradigm.Paradigm({'c': [np.array([0.])]}, [duration],
                                           {'c': [np.array([1.])]})
        s, f, hbr, cbv = phy.create_evoked_physio_signals(phy_params, paradigm,
                                                          ne, tr)
        # shape of a signal: (nb_vox, nb_scans)

        if 0:
            import matplotlib.pyplot as plt
            t = np.arange(f[0].size) * tr
            plt.plot(t, f[0])
            plt.title('inflow')
            plt.show()

        self.assertEqual(s.shape, (paradigm.get_rastered(tr)['c'][0].size,
                                   nb_vox))

        # check signal causality:
        self.assertEqual(f[0, 0], 1.)
        npt.assert_approx_equal(f[-1, 0], 1., significant=2)

        # non-regression test:
        self.assertEqual(np.argmax(f[:, 0]) * tr, 2)

    def test_phy_integrate_euler(self):
        phy_params = phy.PHY_PARAMS_FRISTON00
        tstep = .05
        nb_steps = 400
        stim_duration = int(1 / tstep)
        stim = np.array(
            [1.] * stim_duration + [0.] * (nb_steps - stim_duration))
        epsilon = .5

        s, f, q, v = phy.phy_integrate_euler(phy_params, tstep, stim, epsilon)

        # signal must be causal:
        self.assertEqual(f[0], 1.)
        npt.assert_approx_equal(f[-1], 1., significant=3)

        # non-regression checks:
        npt.assert_approx_equal(np.argmax(f) * tstep, 2.3)
        npt.assert_approx_equal(f.max(), 1.384, significant=4)

        if 0:
            import matplotlib.pyplot as plt
            t = np.arange(nb_steps) * tstep
            plt.plot(t, f)
            plt.title('inflow')
            plt.show()



# -*- coding: utf-8 -*-
import copy as copyModule
import re
import sys
import os
import string
import array
from copy import copy
from inspect import getmro, getsource, currentframe, getargvalues
from inspect import getargspec as _getargspec
from inspect import getouterframes, isfunction

try:
    from collections import OrderedDict
except ImportError:
    from pyhrf.tools.backports import OrderedDict


from xml.dom.minidom import Document
from xml.dom.NodeFilter import NodeFilter


from pyhrf.tools import PickleableStaticMethod

debug = False
debug2 = False
debug3 = False

class FuncWrapper:
    def __init__(self, func, params=None):
        if params is None:
            params = {}
        self.func = func
        self.params = params


def getargspec(func):
    if isinstance(func, PickleableStaticMethod):
        return _getargspec(func.fn)
    else:
        return _getargspec(func)


class XMLable2(object):

    #_init_parameters = None
    #_parameters_value = None

    def __init__(self):

        self._init_parameters = {}

        self._init_func = self.__init__
        #self.check_init_func()
        frame,_,_,_,_,_ = getouterframes(currentframe())[1]
        args, _, _, values = getargvalues(frame)
        if debug:
            print 'self.__class__:', self.__class__
            print 'args:', args
            print 'values:', values
        if 'self' in values:
            values.pop('self')
        for k,v in values.iteritems():
            self._init_parameters[k] = copy(v)
        if debug:
            print 'self._init_parameters:'
            print self._init_parameters

        # args, varargs, varkw, defaults = getargspec(self._init_func)
        # args.remove('self')
        # self._init_parameters = dict( (k,v) for k,v in zip(args, defaults) )



    def check_init_func(self, params=None):

        args, varargs, varkw, defaults = getargspec(self._init_func)
        if varkw is not None: #we don't want **kwargs arguments
            raise Exception('Keywords dict argument (eg **kwargs) ' \
                                'not supported (init function:%s).' \
                                %str(self._init_func))
        if varargs is not None: #we don't want *args arguments
            raise Exception('Positional list argument (eg *args) ' \
                                ' not supported (init function:%s)' \
                                %str(self._init_func))
        if args[0] == 'self' :
            args = args[1:]
        if params is not None:
            if set(params.keys()) != set(args):
                raise Exception('Some arguments do not  '\
                                    'have a value: %s' \
                                    %(str(set(args).difference(params.keys()))))

        elif len(args)-1 != len(defaults):
            pos_args = args[:len(args)-len(defaults)]
            print 'pos_args:', pos_args
            raise Exception('In init function "%s.%s", some arguments are '\
                                'not keywords arguments: %s' \
                                %(str(self.__class__),
                                  str(self._init_func.__name__),
                                  ','.join(pos_args)))


    def set_init_param(self, param_name, param_value):
        if not self._init_parameters.has_key(param_name):
            raise Exception('"%s" is not an argument of init function %s' \
                                %(param_name, self._init_func))
        self._init_parameters[param_name] = copy(param_value)


    def override_param_init(self, init_func,  **params):
        """ TODO (if needed)
        """
        pass

    def set_init(self, init_func, **init_params):

        self._init_func = init_func
        args, varargs, varkw, defaults = getargspec(self._init_func)

        self.check_init_func(init_params)
        self._init_parameters = dict((k,copy(v)) \
                                         for k,v in init_params.iteritems())

        for ip in self._init_parameters:
            if ip not in args:
                raise Exception('Init parameter "%s" is not an argument of '\
                                    'init function "%s". Args are: %s' \
                                    %(ip,str(self._init_func), ','.join(args)))

    def get_parameters_comments(self):
        if debug2: print 'get_parameters_comments ...'
        return getattr(self, 'parametersComments', {})


    def get_parameters_meta(self):
        if debug2: print 'get_parameters_meta ...'
        #TODO
        return {}

    def get_parameters_to_show(self):
        if debug2: print 'get_parameters_to_show ...'
        return getattr(self, 'parametersToShow', [])




class XMLable:

    #_init_parameters = None
    #_parameters_value = None

    def __init__(self,**kwargs):
        #frame = currentframe()
        #if self._init_parameters is None:
        self._init_parameters = {}
        #if self._parameters_value is None:
        self._parameters_value = {}

        if debug:
            print 'XMLable.__init__ (subclass %s) ...' %self.__class__
            print 'self._init_parameters:'
            print self._init_parameters
            print 'self._parameters_value:'
            print self._parameters_value
        frame,_,_,_,_,_ = getouterframes(currentframe())[1]
        args, _, _, values = getargvalues(frame)
        if debug:
            print 'self.__class__:', self.__class__
            print 'args:', args
            print 'values:', values
        if 'self' in values:
            values.pop('self')
        self._init_parameters.update(values.copy())
        if debug:
            print 'self._init_parameters:'
            print self._init_parameters

    def set_init(self, init_func, init_params=None):

        if debug:
            print '~~~~~~~~~~~~~~~~~~~~~'
            print 'set_init ......'
            print 'init_func:', init_params
            print ''

        if init_params is None:

            frame,_,_,_,_,_ = getouterframes(currentframe())[1]
            args, _, _, values = getargvalues(frame)
            # print 'args:', args
            # print 'values:', values
            init_params = dict([(k,values[k]) for k in args if k != 'self'])


        if debug: print 'init_params:', init_params

        self.__init__ = init_func
        self._init_parameters = {}
        self._init_parameters.update(init_params)
        # for k,v in init_params.iteritems():
        #     self._init_parameters[k] = v

    def override_init(self, param_name, init_obj, init_params=None):
        assert not self._parameters_value.has_key(param_name)
        if debug:
            print 'override_init of "%s" with %s for obj of class %s' \
                %(param_name, str(init_obj), self.__class__)
        self._init_parameters[param_name] = FuncWrapper(init_obj, init_params)

    def override_value(self, param_name, value):
        if debug:
            print 'override_value of "%s" with %s for obj of class %s' \
                %(param_name, str(value), self.__class__)
        #assert not self._init_parameters.has_key(param_name)
        self._parameters_value[param_name] = value

    def get_parameters_comments(self):
        return self._gather_parent_dict_parameters('parametersComments')

    def get_parameters_meta(self):
        return self._gather_parent_dict_parameters('parametersMeta')

    def get_parameters_to_show(self):
        if debug2: print 'get_parameters_to_show ...'
        #l = set()
        l = []

        if hasattr(self, "parametersToShow"):
            return self.parametersToShow
        else:
            return []

        # class_tree = getmro(self.__class__)
        # if debug3:
        #     print 'classTree:'
        #     print class_tree
        # for c in class_tree:
        #     if debug: print 'treating class', c
        #     if c != XMLable:
        #         if 'parametersToShow' in getsource(c):
        #             #a bit dirty ...
        #             if debug: print 'found parametersToShow in source ...'
        #             if debug: print 'source:', getsource(c)
        #             p_to_show = getattr(c,'parametersToShow',None)
        #         else:
        #             if debug: print 'did not found parametersToShow '\
        #                     'in source ...'
        #             p_to_show = None
        #         if p_to_show is not None:
        #             for p in p_to_show:
        #                 if p not in l:
        #                     l.append(p)
        #         else: # assume we show everything
        #             print 'show all params ...'
        #             print 'c.__init__:', c.__init__
        #             print '-> argspec:', getargspec(c.__init__)
        #             l.extend(getargspec(c.__init__)[0][1:])
        #             #l.update(getargspec(c.__init__)[0])
        # return l



    def _gather_parent_dict_parameters(self, pname):
        if debug2: print '_gather_parent_dict_parameters ...'
        d = {}
        class_tree = getmro(self.__class__)
        if debug3:
            print 'classTree:'
            print class_tree
        for c in class_tree:
            d.update(getattr(c,pname,{}))
        return d

    def _gather_parent_list_parameters(self, pname):
        if debug2: print '_gather_parent_list_parameters ...'
        l = []
        class_tree = getmro(self.__class__)
        if debug3:
            print 'classTree:'
            print class_tree
        for c in class_tree:
            l.extend(getattr(c,pname,[]))
        return l


class TypedXMLHandler:
    """Class handling the xml format with the following generic document
    structure:

    .. code-block:: none

        <root>
            <tagName 'type'=tagType>
                tagContent
            </tagName>
        </root>

    The root tag is mandatory, so is the 'type' attribute for every other
    tag. This class can parse an xml string and build a dictionary of python
    objects according to each tag (see parseXMLString()). Conversely, it can
    build the xml string corresponding to a list or dictionary of objects (
    see to_xml()). This class is based on the xml.dom python module and
    relies on the DOM structure to parse XML. XML input/output is handled via
    a mapping between a type attribute of a tag and a static handling
    function. This class handles the following basic python types: string,
    int, float, array.array, list, dict. One can add other type-specific read
    or write handlers with the functions addDOMTagReader() and addDOMWriter().

    Reading XML:

    - specific handlers are added with method addDOMTagReader(stype, myReadHandler)
      which maps the function myReadHandler to the string stype.
    - a tag reading handler must have the following prototype:

      .. code-block:: python

        myReadHandler(domTreeWalker):
            # interpret and process tag content
            # return built python object

      , where domTreeWalker is an instance of the
      _xmlplus.dom.TreeWalker.TreeWalker class.
    - useful things to use the domTreeWalker and implement a handler:

      * node = domTreeWalker.currentNode -> current node in the tree parsing,
        of class Node, corresponding to the current tag.
      * node.getAttribute('my attribute') -> return the string corresponding
        to 'my attribute' in the tag definition
      * node.childNodes[0].data -> the tag content data (string type), to be
        parse and build the python object from
      * node.tagName -> the name of the tag
      * node.parentNode -> the parent tag node

    Writing XML:

    - handlers are added with method addDOMTagWriter(pythonType, myWriteHandler),
      where pythonType is of python type 'type' and myWriteHandler a function.
    - a tag writing handler must have the following prototype:

      .. code-block:: python

        myWriteHandler(domDocument, node, pyObj):

      where:

      - domDocument is overall encapsulating structure (_xmlplus.dom.Document instance)
      - node (Node instance) is the current tag node to append data to
      - pyObj is the python object to convert into a 'human-readable' string.
    - useful things to write handlers :
    """
    TYPE_LABEL_NONE = 'none'
    TYPE_LABEL_BOOL = 'bool'
    TYPE_LABEL_STRING = 'char'
    TYPE_LABEL_FLOAT = 'double'
    TYPE_LABEL_INT = 'int'
    TYPE_LABEL_DICT = 'struct'
    TYPE_LABEL_ODICT = 'ordered_struct'
    TYPE_LABEL_LIST = 'list'
    TYPE_LABEL_TUPLE = 'tuple'
    TYPE_LABEL_ARRAY = 'array'
    TYPE_LABEL_KEY_VAL_PAIR = 'dictItem'
    TYPE_LABEL_XML_INCLUDE = 'include'

    ATTRIBUTE_LABEL_PYTHON_CLASS = 'pythonClass'
    ATTRIBUTE_LABEL_PYTHON_CLASS_INIT_MODE = 'pythonInitMode'
    ATTRIBUTE_LABEL_PYTHON_MODULE = 'pythonModule'
    ATTRIBUTE_LABEL_PYTHON_FUNCTION = 'pythonFunction'
    ATTRIBUTE_LABEL_TYPE = 'type'
    ATTRIBUTE_LABEL_META = 'meta'

    def __init__(self, write_callback=None):
        (self.tagDOMReaders, self.objectDOMWriters) = self.packHandlers()

        self.write_callback = write_callback


    def __setstate__(self, dic):
        (dic['tagDOMReaders'], dic['objectDOMWriters']) = self.packHandlers()
        self.__dict__ = dic

    def __getstate__(self): # use for compatibilty with pickles (which doesn't support instance variable methods ...)
        #print self.__class__,'- getstate ... removing static readers/writers ...'
        # copy the __dict__ so that further changes
        # do not affect the current instance
        d = copyModule.deepcopy(self.__dict__)
        # remove the closure that cannot be pickled
        if d.has_key('tagDOMReaders'):
            del d['tagDOMReaders']
        if d.has_key('objectDOMWriters'):
            del d['objectDOMWriters']
        # return state to be pickled
        return d


    def createDocument(self):
        doc = Document()
        root = doc.createElement("root")
        root.setAttribute("pythonXMLHandlerModule",self.__class__.__module__)
        root.setAttribute("pythonXMLHandlerClass",self.__class__.__name__)
        doc.appendChild(root)
        return (doc, root)

    def parseXMLString(self, xml):

        xml = xml.replace('\t', '')
        xml = xml.replace('\n', '')

        # create Reader object
        from xml.dom.ext.reader import Sax2
        reader = Sax2.Reader()

        # parse the document and load it in a DOM tree
        domTree = reader.fromString(xml)

        walker = domTree.createTreeWalker(domTree.documentElement,
                                          NodeFilter.SHOW_ELEMENT, None, 0)

        #print 'walker.currentNode.tagName :', walker.currentNode.tagName
        rootNode = walker.currentNode
        if rootNode.hasAttribute('pythonXMLHandlerModule') \
          and rootNode.hasAttribute('pythonXMLHandlerClass'):
            modulesName = rootNode.getAttribute('pythonXMLHandlerModule')
##            print 'modulesName :', modulesName
            modulesName = modulesName.replace('xmlio.', 'xmliobak.')
            module = eval('__import__("' + modulesName + '", fromlist=[""])')
##            print 'module : ', module
            className = rootNode.getAttribute('pythonXMLHandlerClass')
            xmlHandler = eval('module.'+className+'()')
        else:
            raise Exception('No Typed XML Handler declared in root tag!')
        return xmlHandler.rootTagDOMReader(walker) # An XML document always starts with the root tag

    def buildXMLString(self, obj, label=None, pretty=False):
        (doc,root) = self.createDocument()
        if label is None:
            if hasattr(obj,'xml_label'):
                label = obj.xml_label
            else:
                label = 'anonymObject'

        self.writeDOMData(doc, root, obj, label)
        if not pretty:
            return doc.toxml()
        else:
            return doc.toprettyxml()


    def rootTagDOMReader(self, walker):
        data = {}
        self.parseTerminated = False
        try :
            while(walker.next() and not self.parseTerminated):
                name = walker.currentNode.tagName
                if debug:
                    print '-- name :', name
                data[name] = self.readDOMData(walker)
        except StopIteration:
            pass
        if debug:
            print 'rootTagDOMReader : result :'
            print data
        if len(data)==1:
            return data.values()[0]
        else:
            return data

    def readDOMData(self, walker):
        cn = walker.currentNode
        cnType = cn.getAttribute(self.ATTRIBUTE_LABEL_TYPE)
        to_override = {}
        if debug:
            if debug and 'Prefix' in cn._get_nodeName():
                print 'readDOMData ...'
                print 'cnType:', cnType
                print 'name:', cn._get_nodeName()
                print 'spa:', cn.getAttributeNS('http://www.w3.org/XML/1998/namespace','space')
                if cn.hasAttribute('xml:space'):
                    print 'xml:space found ->', cn.getAttribute('xml:space')
                if cn.hasAttribute('size'):
                    print 'size found'
        # Shade comments and empty childs:
        childsToRemove = []
        preserveSpace = cn.getAttributeNS('http://www.w3.org/XML/1998/namespace',
                                          'space') == "preserve"
        for child in cn.childNodes:

            if (child.nodeType == child.COMMENT_NODE) or \
                    hasattr(child,'data') and str(child.data).strip()=='' and \
                    not preserveSpace:
                childsToRemove.append(child)

        for child in childsToRemove:
            cn.removeChild(child)
        if cn.hasAttribute(self.ATTRIBUTE_LABEL_PYTHON_CLASS) or \
                cn.hasAttribute(self.ATTRIBUTE_LABEL_PYTHON_FUNCTION):
            if debug: print 'building param dict ...'
            paramDict = self.tagDOMReaders[self.TYPE_LABEL_DICT](walker, self)
            for k,v in paramDict.items():
                if isinstance(v, FuncWrapper):
                    if debug: print '%s is FuncWrapper instance' %k
                    paramDict[k] = v.func(**v.params)
                    to_override[k] = v

            #TODO: parse param dict to catch overridings

            #paramDict = dict( [(str(k),v) for k,v in paramDict.iteritems()] )
            if debug:
                print '*** param dict :'
                print paramDict

            #HACK
            if 'BoldEstimationModel' in paramDict:
                paramDict['sampler'] = paramDict.pop('BoldEstimationModel')

            moduleName = cn.getAttribute(self.ATTRIBUTE_LABEL_PYTHON_MODULE)
            module = __import__(moduleName, fromlist=['']) # when fromlist is
                                                           # not empty -> return
                                                           # submodule
##            print 'moduleName = ', moduleName
##            print 'module = ', module
            className = cn.getAttribute(self.ATTRIBUTE_LABEL_PYTHON_CLASS)

            creationMode = cn.getAttribute(self.ATTRIBUTE_LABEL_PYTHON_CLASS_INIT_MODE)
            func_name = cn.getAttribute(self.ATTRIBUTE_LABEL_PYTHON_FUNCTION)
            if debug:
                print 'className:', className
                print 'creationMode:', creationMode
                print 'func_name:', func_name

            if creationMode == 'XMLParamDrivenClass':
                if debug:
                    print 'Building XMLParamDrivenClass ...'
                    print 'to eval :'
                    print 'module.' +className+'(parameters=paramDict, xmlHandler=self, xmlLabel=cn.tagName)'
                try:
                    obj = eval('module.' + className + \
                                   '(parameters=paramDict,' + \
                                   ' xmlHandler=self, xmlLabel=cn.tagName)')
                except TypeError, e:
                    if debug:
                        print 'TypeError while eval ->'
                        print e
                    match_init_args(eval('module.'+className+'()'), paramDict)
                    obj = eval('module.'+className+'(**paramDict)')
            else :
                # Build parameters coma-separated list :
                if debug2:
                    print 'Build parameters coma-separated list :'
                python_ver = sys.version_info
                if python_ver[0] == 2 and python_ver[1] <= 6 \
                      and python_ver[2] <= 4: #<python2.5
                    #print 'fix paramDict!'
                    paramDict = dict([(str(k),v) \
                                          for k,v in paramDict.iteritems()])
                if func_name is not None and func_name != '':
                    if className != '':
                        if debug: print 'func in class !!'
                        if func_name != '__init__':
                            func_str = 'module.' + className + '.' + func_name
                        else:
                            func_str = 'module.' + className
                    else:
                        if debug: print 'func !!'
                        func_str = 'module.' + func_name
                        #return eval('module.' + func_name + '(**paramDict)')
                    #obj = eval('FuncWrapper(%s,paramDict)' %(func_str))
                    if debug: print 'eval(%s(**paramDict))' %(func_str)
                    obj = eval('%s(**paramDict)' %(func_str))
                else:
                    obj = eval('module.'+className+'(**paramDict)')
##        elif cnType == self.TYPE_LABEL_DICT :
##            return self.dictTagDOMReader(walker)

##        elif cnType == self.TYPE_LABEL_LIST :
##            return self.listTagDOMReader(walker)
        elif self.tagDOMReaders.has_key(cnType):
            if debug: print 'using reader from catalog ...'
            obj = self.tagDOMReaders[cnType](walker, self)
        else :
            raise Exception('Unknown tag type :' + cnType)

        if len(to_override) > 0:
            for k,v in to_override.iteritems():
                if debug: print 'override init of %s for obj %s with %s' \
                    %(k,str(obj),str(v.func))
                obj.override_init(k,v.func,v.params)

        if debug:
            print 'readDOMData return:'
            print obj
        return obj

    def writeDOMData(self, doc, node, obj, label, comment=None, meta=None):
        if debug:
            print ''
            print '---------------------------'
            print 'writeDOMData ...'
            print '---------------------------'

        # if override is not None and override.has_key(str(label)):
        #     if debug: print 'overriding %s with %s (original was %s)' \
        #         %(label,str(override[label]),str(obj))
        #     obj = override[label]

        if self.write_callback is not None:
            obj = self.write_callback(obj,label)

        if debug: print ' creating tag with name : ', label
        if debug: print ' type of obj : ', type(obj)
        if debug: print ' meta:', meta
        if hasattr(obj, '__class__'):
            if debug: print ' -> class:', obj.__class__
            if debug: print '---------------------------'
        newNode = doc.createElement(label)
        if comment is not None:
            com = doc.createComment(comment)
            newNode.appendChild(com)
        if meta is not None:
            newNode.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_META,meta)
##        if type(obj) == type({}):
##            self.dictDOMWriter(doc, newNode, obj)
##        elif type(obj) == type([]):
##            self.listDOMWriter(doc, newNode, obj)

        if self.inspectable(obj):
            if debug: print 'inspectable ...'
            self.inspect_and_append_to_DOM_tree(doc, newNode, obj)
        elif self.objectDOMWriters.has_key(type(obj)):
            if debug: print '  calling handler ...'
            self.objectDOMWriters[type(obj)](doc, newNode, obj, self)
        elif hasattr(obj, 'appendParametersToDOMTree'):
            if debug: print '  calling interface method ...'
            obj.appendParametersToDOMTree(doc, newNode)
        else:
            error = 'no xml policy to dump item!\n'
            error += ' tag name was : %s\n' %label
            error += ' type of obj was : %s\n' %str(type(obj))
            raise Exception(error)


        #TODO else: Exception ...
        node.appendChild(newNode)

    def inspectable(self, obj):
        if debug:
            print 'func inspectable? ....'
            print 'obj:', obj
            print 'obj class:', obj.__class__
        #print obj
        if isinstance(obj, FuncWrapper):
            if debug: print '-> FuncWrapper!'
            obj_init = obj.func
        elif isfunction(obj):
            if debug: print '-> function!'
            obj_init = obj
        else:
            if hasattr(obj, 'im_func'): #classmethod
                if debug: print '-> im_func found'
                obj_init = obj.im_func
            elif isinstance(obj, XMLable): # assume instance
                if debug: print '-> instance obj !!!!!'
                obj_init = obj.__init__
            elif isinstance(obj, XMLable2): # assume instance
                if debug: print '-> XMLable2 instance obj !!!!!'
                obj_init = obj._init_func
            else:
                if debug: print '-> simple instance ...'
                return False

        if debug: print 'obj_init:', obj_init
        a = getargspec(obj_init)
        if debug:
            print 'argspec:', a
        if len(a[0]) == 0:
            return True
        if debug:
            print 'a.keywords is None:', a[2] is None
            print 'a.varargs is None', a[1] is None
            print 'len(a.args)', len(a[0])
            if a[3] is not None:
                print 'len(a.defaults)', len(a[3])
            print 'a:', a

        if a[3] is None:
            raise Exception("No default parameters for function %s"\
                                %(obj_init.__name__))

        # no keywords, no varargs,  all parameters have default values
        return a[2] is None and a[1] is None and \
            (len(a[0])-1 == len(a[3]))


    def inspect_and_append_to_DOM_tree(self, doc, node, obj):
        override_for_func = None
        if debug:
            print 'inspect_and_append_to_DOM_tree ...'
            print 'Test if obj is a function :'

        if isinstance(obj, FuncWrapper):
            if debug: print '-> FuncWrapper instance  detected'
            node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_MODULE,
                              obj.func.__module__)
            node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_FUNCTION,
                              obj.func.__name__)
            if hasattr(obj.func, 'im_self'):
                if debug: print '-> class method  detected'
                node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_CLASS,
                                  obj.func.im_self.__name__)

            obj_init = obj.func
            override_for_func = obj.params
        elif isfunction(obj) or hasattr(obj, 'im_func'):

            node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_MODULE,
                              obj.__module__)
            node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_FUNCTION,
                              obj.__name__)

            if hasattr(obj, 'im_self'):
                if debug: print '-> class method  detected'
                node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_CLASS,
                                  obj.im_self.__name__)
                obj_init = obj.im_func
            else:
                if debug: print '-> function  detected'
                obj_init = obj
        elif isinstance(obj, XMLable):
            if debug: print 'XMLable -> instance  detected'
            obj_init = obj.__init__
            node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_CLASS,
                              obj.__class__.__name__)
            node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_MODULE,
                              obj.__class__.__module__)

        elif isinstance(obj, XMLable2):
            if debug: print '-> XMLable2 instance  detected'
            obj_init = obj._init_func
            if isinstance(obj_init, PickleableStaticMethod):
                obj_init = obj_init.fn
            node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_CLASS,
                              obj.__class__.__name__)
            node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_MODULE,
                              obj.__class__.__module__)
            if obj_init != obj.__init__:
                attr_name = TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_FUNCTION
                try:
                    node.setAttribute(attr_name, obj_init.im_func.__name__)
                except AttributeError:
                    node.setAttribute(attr_name, obj_init.__name__)
        else:
            raise Exception("Unsupported obj: %s", str(obj))

        node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_TYPE,
                          TypedXMLHandler.TYPE_LABEL_DICT)

        p_comments = {}
        p_to_show = []
        p_meta = {}
        if issubclass(obj.__class__, XMLable) or \
                issubclass(obj.__class__, XMLable2):
            if debug: print 'instance is of XMLable(2) subclass'
            p_comments = obj.get_parameters_comments()
            p_to_show = obj.get_parameters_to_show()
            p_meta = obj.get_parameters_meta()
            if debug:
                print 'p_comments:', p_comments
                print 'p_to_show:', p_to_show
                print 'p_meta:', p_meta

        if debug: print 'Function used for init:', obj_init
        a = getargspec(obj_init)
        if debug: print 'argspec:', a
        if len(a[0]) == 0:
            return

        if len(p_to_show) == 0:
            if debug: print 'p_to_show is empty ...'
            arg_list = a[0][1:]
            arg_defaults = a[3]
        else:
            if debug: print 'use p_to_show ...', p_to_show
            arg_list = []
            arg_defaults = []
            for p in p_to_show:
                if p in a[0][1:]:
                    arg_list.append(p)
                    arg_defaults.append(a[3][a[0][1:].index(p)])


        if debug:
            print 'arg_list :'
            print arg_list
            if hasattr(obj, '_init_parameters'):
                print 'obj._init_parameters:'
                print obj._init_parameters

        if debug:
            print 'Parsing args and wrinting them to XML ...'
        for arg, default in zip(arg_list,arg_defaults):
            override = None
            if debug:
                print 'arg:', arg
                #print 'default:', default
            comment = p_comments.get(arg,None)
            meta = p_meta.get(arg,None)

            if override_for_func is not None:
                if debug: print '-> override arg values provided'
                default = override_for_func.get(arg,default)
            if hasattr(obj, '_parameters_value') and \
                    obj._parameters_value is not None and \
                    obj._parameters_value.has_key(arg):
                if debug: print '-> arg provided in obj._parameters_value'
                #print obj._parameters_value
                default = obj._parameters_value.get(arg)
            elif hasattr(obj, '_init_parameters') and \
                    obj._init_parameters is not None and \
                    obj._init_parameters.has_key(arg):
                if debug:
                    print '-> init func for arg provided in ' \
                        'obj._init_parameters'
                default = obj._init_parameters.get(arg)

            if debug2:
                print 'default:', default
            self.writeDOMData(doc, node, default, arg, comment, meta)


    def packHandlers(self):
        tagDOMReaders = {}
        objectDOMWriters = {}

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_DICT] = TypedXMLHandler.dictTagDOMReader
        objectDOMWriters[dict] = TypedXMLHandler.dictDOMWriter


        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_ODICT] = TypedXMLHandler.odictTagDOMReader
        objectDOMWriters[OrderedDict] = TypedXMLHandler.odictDOMWriter


        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_LIST] = TypedXMLHandler.listTagDOMReader
        objectDOMWriters[list] = TypedXMLHandler.listDOMWriter

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_TUPLE] = TypedXMLHandler.tupleTagDOMReader
        objectDOMWriters[tuple] = TypedXMLHandler.tupleDOMWriter

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_STRING] = TypedXMLHandler.stringTagDOMReader
        objectDOMWriters[str] = TypedXMLHandler.stringDOMWriter
        objectDOMWriters[unicode] = TypedXMLHandler.stringDOMWriter

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_FLOAT] = TypedXMLHandler.floatTagDOMReader
        objectDOMWriters[float] = TypedXMLHandler.floatDOMWriter

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_INT] = TypedXMLHandler.intTagDOMReader
        objectDOMWriters[int] = TypedXMLHandler.intDOMWriter

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_BOOL] = TypedXMLHandler.boolTagDOMReader
        objectDOMWriters[bool] = TypedXMLHandler.boolDOMWriter

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_ARRAY] = TypedXMLHandler.arrayTagDOMReader
        objectDOMWriters[type(array.array('i'))] = TypedXMLHandler.arrayDOMWriter

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_NONE] = TypedXMLHandler.noneTagDOMReader
        objectDOMWriters[type(None)] = TypedXMLHandler.noneDOMWriter

        tagDOMReaders[TypedXMLHandler.TYPE_LABEL_XML_INCLUDE] = TypedXMLHandler.includeTagDOMReader

        return (tagDOMReaders, objectDOMWriters)

    def mountDefaultHandlers(self):
        (self.tagDOMReaders, self.objectDOMWriters) = self.packHandlers()

    # xml I/O for 'char' type (python string basic type):
    def noneTagDOMReader(walker, xmlHandler):
        return None
    noneTagDOMReader = staticmethod(noneTagDOMReader)


    def noneDOMWriter(doc, node, noneObj, xmlHandler):
        node.setAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE, xmlHandler.TYPE_LABEL_NONE)
    noneDOMWriter = staticmethod(noneDOMWriter)

    def includeTagDOMReader(walker, xmlHandler):
        includeFn = walker.currentNode.childNodes[0].data
        assert os.path.exists(includeFn)
        f = open(includeFn, 'r')
        includeContent = string.join(f.readlines())
        f.close()
        return from_xml(includeContent)

    includeTagDOMReader = staticmethod(includeTagDOMReader)


    # xml I/O for 'char' type (python string basic type):
    def stringTagDOMReader(walker, xmlHandler):
        if len(walker.currentNode.childNodes) == 0:
            return ''
        else:
            return str(walker.currentNode.childNodes[0].data)
    stringTagDOMReader = staticmethod(stringTagDOMReader)


    def stringDOMWriter(doc, node, stringObj, xmlHandler):
        node.setAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE,
                          xmlHandler.TYPE_LABEL_STRING)
        size = str(len(stringObj))
        node.setAttribute('size', size)
        if debug:
            print 'createTextNode, from string:', stringObj
            print 'size:', size
        textData = doc.createTextNode(stringObj)
        if debug: print 'textData:', textData
        node.appendChild(textData)
    stringDOMWriter = staticmethod(stringDOMWriter)

    # xml I/O for 'bool' type:
    def boolTagDOMReader(walker, xmlHandler):
        return bool(int(walker.currentNode.childNodes[0].data))
    boolTagDOMReader = staticmethod(boolTagDOMReader)

    def boolDOMWriter(doc, node, boolObj, xmlHandler):
        node.setAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE,
                          xmlHandler.TYPE_LABEL_BOOL)
        node.setAttribute('size', '1')
        textData = doc.createTextNode(str(int(boolObj)))
        node.appendChild(textData)
    boolDOMWriter = staticmethod(boolDOMWriter)

    # xml I/O for 'float' type :
    def floatTagDOMReader(walker, xmlHandler):
        return float(walker.currentNode.childNodes[0].data)
    floatTagDOMReader = staticmethod(floatTagDOMReader)


    def floatDOMWriter(doc, node, floatObj, xmlHandler):
        node.setAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE, xmlHandler.TYPE_LABEL_FLOAT)
        node.setAttribute('size', '1')
        textData = doc.createTextNode(str(floatObj))
        node.appendChild(textData)
    floatDOMWriter = staticmethod(floatDOMWriter)

    # xml I/O for 'int' type :
    def intTagDOMReader(walker, xmlHandler):
        return int(walker.currentNode.childNodes[0].data)
    intTagDOMReader = staticmethod(intTagDOMReader)


    def intDOMWriter(doc, node, intObj, xmlHandler):
        node.setAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE, xmlHandler.TYPE_LABEL_INT)
        node.setAttribute('size', '1')
        textData = doc.createTextNode(str(intObj))
        node.appendChild(textData)
    intDOMWriter = staticmethod(intDOMWriter)

    # xml I/O for 'array' type :
    def arrayTagDOMReader(walker, xmlHandler):
        return int(walker.currentNode.childNodes[0].data)
    arrayTagDOMReader = staticmethod(arrayTagDOMReader)


    def arrayDOMWriter(doc, node, arrayObj, xmlHandler):
        node.setAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE, xmlHandler.TYPE_LABEL_ARRAY)
        node.setAttribute('size', str(len(array)))
        node.setAttribute('subtype', arrayObj.typecode)
        textData = doc.createTextNode(map(str,arrayObj)) ## check map(...) -> list de string & createTextNode(arg=string) ??
        node.appendChild(textData)
    arrayDOMWriter = staticmethod(arrayDOMWriter)

    # xml I/O for 'list' type:
    def listTagDOMReader(walker, xmlHandler):
        currentListNode = walker.currentNode
        #result = range(int(currentListNode.getAttribute('size')))
        result = dict()
        while 1:
            nn = walker.nextNode()
            if nn==None:
                xmlHandler.parseTerminated = True
                break

            if not nn.parentNode.isSameNode(currentListNode) :
                walker.previousNode()
                break
            result[int(nn.tagName[1:])] = xmlHandler.readDOMData(walker)
        return [result[R] for R in sorted(result.keys())]
    listTagDOMReader = staticmethod(listTagDOMReader)

    def listDOMWriter(doc, node, listObj, xmlHandler):
        node.setAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE, xmlHandler.TYPE_LABEL_LIST)
        node.setAttribute( 'size', str(len(listObj)) )
        i = 0
        for elem in listObj :
            xmlHandler.writeDOMData(doc, node, elem, 'i'+str(i))
            i += 1
    listDOMWriter = staticmethod(listDOMWriter)



    # xml I/O for 'list' type:
    def tupleTagDOMReader(walker, xmlHandler):
        currentListNode = walker.currentNode
        result = range(int(currentListNode.getAttribute('size')))
        while 1:
            nn = walker.nextNode()
            if nn==None:
                xmlHandler.parseTerminated = True
                break

            if not nn.parentNode.isSameNode(currentListNode) :
                walker.previousNode()
                break
            result[int(nn.tagName[1:])] = xmlHandler.readDOMData(walker)
        return tuple(result)
    tupleTagDOMReader = staticmethod(tupleTagDOMReader)

    def tupleDOMWriter(doc, node, tupleObj, xmlHandler):
        node.setAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE, xmlHandler.TYPE_LABEL_TUPLE)
        node.setAttribute( 'size', str(len(tupleObj)) )
        i = 0
        for elem in tupleObj :
            xmlHandler.writeDOMData(doc, node, elem, 'i'+str(i))
            i += 1
    tupleDOMWriter = staticmethod(tupleDOMWriter)



    # xml I/O for 'dict' type:
    def dictTagDOMReader(walker, xmlHandler, init_class=None):
        if debug:
            print 'dictTagDOMReader ...'
        currentDictNode = walker.currentNode
        if debug:
            print ' currentNode : ',currentDictNode.tagName
        if init_class is None:
            result = {}
        else:
            result = init_class()

        if debug:
            print 'dict init:', result.__class__

        while 1:
            nn = walker.nextNode()
            if nn==None:
                xmlHandler.parseTerminated = True
                break
            if not nn.parentNode.isSameNode(currentDictNode) :
                walker.previousNode()
                break
            nnType = nn.getAttribute(xmlHandler.ATTRIBUTE_LABEL_TYPE)
            if nnType == xmlHandler.TYPE_LABEL_KEY_VAL_PAIR :
                walker.nextNode()
                key = xmlHandler.readDOMData(walker)
                walker.nextNode()
                result[key] = xmlHandler.readDOMData(walker)
            else :
                if debug: print 'nn.tagName:', nn.tagName
                result[nn.tagName] = xmlHandler.readDOMData(walker)

        if debug:
            print ' returning ...:', result
        return result
    dictTagDOMReader = staticmethod(dictTagDOMReader)

    def dictDOMWriter(doc, node, dictObj, xmlHandler, atype=None):

        if atype is None:
            atype = xmlHandler.TYPE_LABEL_DICT

        atypel = xmlHandler.ATTRIBUTE_LABEL_TYPE
        node.setAttribute(atypel, atype)
        for key,val in dictObj.items() :
            # if the key type is string, use the standard xml
            # association : string label <-> string data
            # if key starts with a digit -> tag name will not be well-formed
            if (type(key)==type('') or type(key)==unicode)\
                   and not key[0].isdigit():
                xmlHandler.writeDOMData(doc, node, val, key)
            else : # if the key type is not a string,
                   # build a key<->value pair xml structure
                keyValElem = doc.createElement("anonym")
                atypel = xmlHandler.ATTRIBUTE_LABEL_TYPE
                atype = xmlHandler.TYPE_LABEL_KEY_VAL_PAIR
                keyValElem.setAttribute(atypel, atype)
                xmlHandler.writeDOMData(doc, keyValElem, key, "key")
                xmlHandler.writeDOMData(doc, keyValElem, val, "value")
                node.appendChild(keyValElem)
    dictDOMWriter = staticmethod(dictDOMWriter)

    def odictTagDOMReader(walker, xmlHandler):
        return xmlHandler.dictTagDOMReader(walker, xmlHandler, OrderedDict)
    odictTagDOMReader = staticmethod(odictTagDOMReader)


    def odictDOMWriter(doc, node, dictObj, xmlHandler):
        return xmlHandler.dictDOMWriter(doc, node, dictObj, xmlHandler,
                                        xmlHandler.TYPE_LABEL_ODICT)
    odictDOMWriter = staticmethod(odictDOMWriter)

def from_xml(s, handler = TypedXMLHandler()):
    return handler.parseXMLString(s)


# def loadXML(fxml):
#     f = open(fxml, 'r')
#     sXml = string.join(f.readlines())
#     f.close()
#     return from_xml(sXml)

def to_xml(o, handler = TypedXMLHandler(), objName="anonymObject", pretty=False):
    if isinstance(o, XMLParamDrivenClass) or hasattr(o, 'xmlHandler'):
        handler = o.xmlHandler
    return handler.buildXMLString(o, objName, pretty=pretty)



def read_xml(fn):
    f = open(fn)
    content = f.read()
    f.close()
    from xmlnumpy import NumpyXMLHandler
    handler = NumpyXMLHandler()
    return handler.parseXMLString(content)

def write_xml(obj, fn):
    fOut = open(fn, 'w')
    from xmlnumpy import NumpyXMLHandler
    fOut.write(to_xml(obj, handler=NumpyXMLHandler()))
    fOut.close()

class XMLParamDrivenClassInitException :

    def __init__(self):
        pass


class XMLParamDrivenClass:
    """
    Base \"abstract\" class to handle parameters with clear specification and default values.
    Recursive aggregation is availaible to handle aggregated variables which also require
    parameter specifications.
    """

    # Parameters labels should be clearly defined in derived classes, example :
    # P_PARAMETER_1 = 'labelParameter1'
    # P_PARAMETER_1 would then be the inner designation and 'labelParameter1' would be used in the outter XML specification.

    # Default parameter values :
    defaultParameters = {}

    # Comments describing parameters:
    parametersComments = {}

    # parameters meta type:
    parametersMeta = {}

    def __init__(self, parameters=None, xmlHandler=TypedXMLHandler(),
                 xmlLabel=None, xmlComment=None):
        """
        Create a new XMLParamDrivenClass

        """
        self.updateParameters(parameters)
        self.xmlHandler = xmlHandler

        if xmlLabel is None :
            # Default tag label used in xml
            self.xmlLabel = self.__class__.__name__+"_parameters"
        else :
            self.xmlLabel = xmlLabel

        if xmlComment is None:
            #self.xmlComment = 'Set of parameters defining the python class : ' \
            #                  +self.__module__ +'.'+ self.__class__.__name__
            self.xmlComment = None
        else :
            self.xmlComment = xmlComment

    def fetchDefaultParameters(self):
        self.parameters = copyModule.deepcopy(self.defaultParameters)

    def updateParameters(self, newp):
        self.fetchDefaultParameters()

        # Extracting new parameters' values from input :
        if newp != None :
            for key in newp.keys():
                if self.parameters.has_key(key):
                    self.parameters[key] = newp[key]
                else :
                    print 'Warning - in creation of a \"'+self.__class__.__name__ +'\" instance :'
                    print '  unknown parameter designation : ', key

    def appendParametersToDOMTree(self, doc, node):
#         if tagName==None:
#             tag = self.xmlLabel
#         else :
#             tag = tagName
#         elem = doc.createElement(tag)
        node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_CLASS,
                          self.__class__.__name__)
        node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_PYTHON_MODULE,
                          self.__class__.__module__)
        node.setAttribute(self.xmlHandler.ATTRIBUTE_LABEL_PYTHON_CLASS_INIT_MODE,
                          "XMLParamDrivenClass")

        node.setAttribute(TypedXMLHandler.ATTRIBUTE_LABEL_TYPE,
                          TypedXMLHandler.TYPE_LABEL_DICT)

        if self.xmlComment is not None:
            com = doc.createComment(self.xmlComment)
            node.appendChild(com)

        if hasattr(self, 'parametersToShow'):
            paramIter = self.parametersToShow
        else:
            paramIter = self.parameters.keys()

        for paramLabel in paramIter:
            paramVal = self.parameters[paramLabel]
            if debug: print 'treating parameter : ',paramLabel
            if debug: print 'val : ',paramVal
            #print 'self.xmlHandler :', self.xmlHandler
            comment = self.parametersComments.get(paramLabel,None)
            meta = self.parametersMeta.get(paramLabel)
            self.xmlHandler.writeDOMData(doc, node, paramVal, paramLabel,
                                         comment=comment, meta=meta)
        #node.appendChild(elem)

    def parametersToXml(self, tagName=None, pretty=False):
        (doc,root) = self.xmlHandler.createDocument()
        if tagName==None:
            tag = self.xmlLabel
        else :
            tag = tagName
        node = doc.createElement(tag)
        self.appendParametersToDOMTree(doc, node)
        root.appendChild(node)
        if not pretty:
            return doc.toxml()
        else:
            return doc.toprettyxml()


def match_init_args(c, argsDict):

    a = getargspec(c.__init__)
    to_pop = []
    for argname in argsDict.iterkeys():
        if argname not in a[0]:
            print 'Warning: %s unused for class %s' \
                %(argname, c.__class__)
            to_pop.append(argname)

    for tp in to_pop:
        argsDict.pop(tp)



# -*- coding: utf-8 -*-

import logging

import numpy as np

from pyhrf.jde.samplerbase import *
from pyhrf import xmlio


logger = logging.getLogger(__name__)


class NoiseVariance_Drift_Sampler(xmlio.XmlInitable, GibbsSamplerVariable):

    def __init__(self, do_sampling=True, use_true_value=False,
                 val_ini=None):
        """
        #TODO : comment
        """
        xmlio.XmlInitable.__init__(self)
        sampleFlag = do_sampling
        valIni = val_ini
        useTrueVal = use_true_value

        GibbsSamplerVariable.__init__(self, 'noise_var', valIni=valIni,
                                      useTrueValue=useTrueVal,
                                      sampleFlag=sampleFlag, axes_names=[
                                          'voxel'],
                                      value_label='PM Noise Var')

    def linkToData(self, dataInput):
        self.dataInput = dataInput
        self.nbConditions = self.dataInput.nbConditions
        self.nbVox = self.dataInput.nbVoxels
        self.ny = self.dataInput.ny

    def checkAndSetInitValue(self, variables):
        if self.currentValue is None:
            if not self.sampleFlag and self.dataInput.simulData != None:
                self.currentValue = self.dataInput.simulData.noise.variance
            else:
                self.currentValue = 0.5 * self.dataInput.varData

    def sampleNextInternal(self, variables):

        # print 'Step 3 : Noise variance Sampling *****RelVar*****'
        # print '         varYbar begin =',self.get_variable('nrl').varYbar.sum()
        # print '         varYtilde begin
        # =',self.get_variable('nrl').varYtilde.sum()

        var_y_bar = self.get_variable('nrl').varYbar

        beta = (var_y_bar * var_y_bar).sum(0) / 2.

        # gammaSamples = np.random.gamma(0.5*(self.ny - self.dataInput.colP +1)-1, 1,
        #                            self.nbVox)

        #gammaSamples = np.random.gamma((self.ny + 1.)/2, 1, self.nbVox)
        gammaSamples = np.random.gamma(
            (self.ny) / 2., 1, self.nbVox)  # Jeffrey's prior

        np.divide(beta, gammaSamples, self.currentValue)


class NoiseVarianceSampler(xmlio.XmlInitable, GibbsSamplerVariable):
    """
    #TODO : comment

    """

    def __init__(self, do_sampling=True, use_true_value=False,
                 val_ini=None):

        xmlio.XmlInitable.__init__(self)
        GibbsSamplerVariable.__init__(self, 'noise_var', valIni=val_ini,
                                      sampleFlag=do_sampling,
                                      useTrueValue=use_true_value,
                                      axes_names=['voxel'],
                                      value_label='PM Noise Var')

    def linkToData(self, dataInput):
        self.dataInput = dataInput
        self.nbConditions = self.dataInput.nbConditions
        self.nbVox = self.dataInput.nbVoxels
        self.ny = self.dataInput.ny
        self.nbColX = self.dataInput.nbColX

        # Do some allocations :
        self.beta = np.zeros(self.nbVox, dtype=float)
        self.mXhQXh = np.zeros((self.nbConditions, self.nbConditions),
                               dtype=float)

        if self.dataInput.simulData is not None:
            if isinstance(self.dataInput.simulData, dict):
                if dataInput.simulData.has_key('v_gnoise'):
                    self.trueValue = self.dataInput.simulData['v_noise']
                    logger.info('True noise variance = %1.3f', self.trueValue)

            elif isinstance(dataInput.simulData, list):
                sd = dataInput.simulData[0]
                if isinstance(sd, dict):
                    self.trueValue = sd['noise'].var(0).astype(np.float64)
                else:
                    self.trueValue = sd.noise.variance
            else:
                self.trueValue = self.dataInput.simulData.noise.variance

    def checkAndSetInitValue(self, variables):
        if self.trueValue is not None and np.isscalar(self.trueValue):
            self.trueValue = self.trueValue * np.ones(self.nbVox)
        if self.currentValue == None:
            if self.useTrueValue:
                if self.trueValue is not None:
                    logger.info('Use true noise variance value ...')
                    self.currentValue = self.trueValue
                else:
                    raise Exception(
                        'True noise varaince have to be used but none defined.')
            else:
                self.currentValue = 0.5 * self.dataInput.varData

        # if self.currentValue == None :
            # if not self.sampleFlag and self.dataInput.simulData != None :
                #self.currentValue = self.dataInput.simulData.noise.variance
            # else:
                #self.currentValue = 0.1 * self.dataInput.varData

    def computeMXhQXh(self, h, varXQX):
        for j in xrange(self.nbConditions):
            for k in xrange(self.nbConditions):
                self.mXhQXh[j, k] = np.dot(np.dot(h, varXQX[j, k]), h)

    def compute_aaXhQXhi(self, aa, i):
        aaXhQXhi = 0.0
        for j in xrange(self.nbConditions):
            for k in xrange(self.nbConditions):
                aaXhQXhi += aa[j, k, i] * self.mXhQXh[j, k]
                # print 'aa[%d,%d,%d] * self.mXhQXh[%d,%d] = %f * %f = %f' \
                #    %(j,k,i,j,k, aa[j,k,i], self.mXhQXh[j,k],
                #      aa[j,k,i] * self.mXhQXh[j,k])
                # print aa[j,k,i] * self.mXhQXh[j,k],', ',
                # print ''
                #aaXhQXhiTab[i] += snrl.aa[j,k,i] * self.mXhQXh[j,k]
        return aaXhQXhi

    def sampleNextInternal_bak(self, variables):

        y_tilde = self.samplerEngine.get_variable('nrl').varYtilde
        beta = (y_tilde * y_tilde).sum(0) / 2
        gammaSamples = np.random.gamma((self.ny - 1.) / 2, 1, self.nbVox)

        np.divide(beta, gammaSamples, self.currentValue)

    def sampleNextInternal(self, variables):
        # TODO : comment

        h = self.get_variable('hrf').currentValue
        varXQX = self.dataInput.matXQX
        snrl = self.get_variable('nrl')

        saXh = snrl.sumaXh
        # Example to retrieve variable from sharedData :
        # saXh = self.samplerEngine.sharedData.get('varXh')

        yTQy = self.dataInput.yTQy
        matQy = self.dataInput.matQy

        self.computeMXhQXh(h, varXQX)

        aa = snrl.aa.swapaxes(2, 0)
        aaXhQXh = (aa * self.mXhQXh).sum(1).sum(1)
        if 0:
            for i in xrange(self.nbVox):
                self.beta[i] = .5 * (yTQy[i] - 2 * np.dot(saXh[:, i], matQy[:, i])
                                     + aaXhQXh[i])

        else:
            y_tilde = self.samplerEngine.get_variable('nrl').varYtilde
            for i in xrange(self.nbVox):
                varYtildeTdelta = np.dot(y_tilde[:, i], self.dataInput.delta)
                self.beta[i] = 0.5 * np.dot(varYtildeTdelta, y_tilde[:, i])
        logger.debug('All betas apost :')
        logger.debug(self.beta)
        logger.info('betas apost = %1.3f(%1.3f)', self.beta.mean(),
                    self.beta.std())

        gammaSamples = np.random.gamma(0.5 * (self.ny - self.dataInput.colP - 1),
                                       1, self.nbVox)

        logger.debug('sigma2 ~betas/Ga(%1.3f,1)',
                     0.5 * (self.ny - self.dataInput.colP + 1))
        self.currentValue = np.divide(self.beta, gammaSamples)

        logger.info('noise vars = %1.3f(%1.3f)', self.currentValue.mean(),
                    self.currentValue.std())
        if (self.currentValue < 0).any():
            print 'negative variances !!'
            print 'at vox : ', np.where(self.currentValue < 0)
            print '-> betas = ', self.beta[np.where(self.currentValue < 0)]
            print '-> gamma samples = ', gammaSamples[np.where(self.currentValue < 0)]
            print '-> yTQy[i] = ', yTQy[np.where(self.currentValue < 0)]
            i = np.where(self.currentValue < 0)[0]
            print '-> setting to almost zero'
            self.currentValue[np.where(self.currentValue < 0)] = 0.0001

    def finalizeSampling(self):
        GibbsSamplerVariable.finalizeSampling(self)
        del self.mXhQXh
        del self.beta

#######################################################
# Noise Variance sampler with Relevant Variable
#######################################################


class NoiseVarianceSamplerWithRelVar(NoiseVarianceSampler):

    def compute_aawwXhQXhi(self, ww, aa, i):
        aawwXhQXhi = 0.0
        for j in xrange(self.nbConditions):
            for k in xrange(self.nbConditions):
                aawwXhQXhi += ww[j, k] * aa[j, k, i] * self.mXhQXh[j, k]
        return aawwXhQXhi

    def computeWW(self, w, destww):
        for j in xrange(self.nbConditions):
            for k in xrange(self.nbConditions):
                destww[j, k] = w[j] * w[k]

    def sampleNextInternal(self, variables):
        # TODO : comment

        h = self.get_variable('hrf').currentValue
        varXQX = self.dataInput.matXQX
        snrl = self.get_variable('nrl')
        w = self.get_variable('W').currentValue

        swaXh = snrl.sumWaXh

        nrl = snrl.currentValue
        varYtilde = snrl.varYtilde
        yTQy = self.dataInput.yTQy
        matQy = self.dataInput.matQy

        self.computeMXhQXh(h, varXQX)

        ww = np.zeros((self.nbConditions, self.nbConditions), dtype=int)
        self.computeWW(w, ww)

        aa = snrl.aa.swapaxes(2, 0)
        aaww = aa * ww
        aawwXhQXh = (aaww * self.mXhQXh).sum(1).sum(1)

        for i in xrange(self.nbVox):
            self.beta[i] = .5 * (yTQy[i] - 2 * np.dot(swaXh[:, i], matQy[:, i])
                                 + aawwXhQXh[i])

        logger.debug('All betas apost :')
        logger.debug(self.beta)
        logger.info('betas apost = %1.3f(%1.3f)', self.beta.mean(),
                    self.beta.std())

        gammaSamples = np.random.gamma(0.5 * (self.ny - self.dataInput.colP + 1) - 1, 1,
                                       self.nbVox)

        logger.debug('sigma2 ~betas/Ga(%1.3f,1)',
                     0.5 * (self.ny - self.dataInput.colP + 1))
        self.currentValue = np.divide(self.beta, gammaSamples)

        logger.info('noise vars = %1.3f(%1.3f)', self.currentValue.mean(),
                    self.currentValue.std())
        if (self.currentValue < 0).any():
            print 'negative variances !!'
            print 'at vox : ', np.where(self.currentValue < 0)
            print '-> betas = ', self.beta[np.where(self.currentValue < 0)]
            print '-> gamma samples = ', gammaSamples[np.where(self.currentValue < 0)]
            print '-> yTQy[i] = ', yTQy[np.where(self.currentValue < 0)]
            i = np.where(self.currentValue < 0)[0]
            print '-> np.dot(swaXh[:,i].transpose(), matQy[:,i]) =', np.dot(swaXh[:, i].transpose(), matQy[:, i])

    def finalizeSampling(self):
        GibbsSamplerVariable.finalizeSampling(self)
        del self.mXhQXh
        del self.beta


#######################################################
# Noise Variance sampler with Habituation
#######################################################


# class NoiseVariancewithHabSampler(NoiseVarianceSampler):
    # pass

    #"""--------------------

class NoiseVariancewithHabSampler(NoiseVarianceSampler):
    """
    #TODO : Sampling procedure for noise variance parameters (white noise)
    #in case of habituation modeling wrt magnitude

    """

    def sampleNextInternal(self, variables):
        # TODO : comment
        h = self.get_variable('hrf').currentValue
        smplNRL = self.get_variable('nrl')
        aXh = smplNRL.aXh
        saXh = smplNRL.sumaXh
        varYtilde = smplNRL.varYtilde
        # Example to retrieve variable from sharedData :
        # saXh = self.samplerEngine.sharedData.get('varXh')

        yTQy = self.dataInput.yTQy
        matQy = self.dataInput.matQy

        for i in xrange(self.nbVox):
            aaXhQXhi = 0.0
            for j in xrange(self.nbConditions):
                for k in xrange(self.nbConditions):
                    aaXhQXhi += np.dot(np.dot(aXh[:, i, j],
                                              self.dataInput.delta), aXh[:, i, k])

            self.beta[i] = .5 * (yTQy[i] - 2 * np.dot(saXh[:, i], matQy[:, i])
                                 + aaXhQXhi)

        logger.debug('All betas apost :')
        logger.debug(self.beta)
        logger.debug('betas apost = %1.3f(%1.3f)', self.beta.mean(),
                     self.beta.std())

        a = 0.5 * (self.ny - self.dataInput.colP + 1)
        gammaSamples = np.random.gamma(a, 1, self.nbVox)

        logger.debug('sigma2 ~betas/Ga(%1.3f,1)', a)
        self.currentValue = np.divide(self.beta, gammaSamples)

        logger.debug('All noise vars :')
        logger.debug(self.currentValue)
        logger.info('noise vars = %1.3f(%1.3f)', self.currentValue.mean(),
                    self.currentValue.std())

    def finalizeSampling(self):

        del self.beta


#######################################################
# Noise Variance AR sampler
#######################################################
class NoiseVarianceARSampler(NoiseVarianceSampler):

    def checkAndSetInitValue(self, variables):

        NoiseVarianceSampler.checkAndSetInitValue(self, variables)
        self.varYTilde = np.empty((self.ny, self.nbVox), dtype=float)

    def computeVarYTilde(self, varNrls, varXh, varMBYPl):
        for v in xrange(self.nbVox):
            repNRLv = np.tile(varNrls[:, v], (self.ny, 1))
            avjXjh = repNRLv * varXh
            self.varYTilde[:, v] = varMBYPl[:, v] - avjXjh.sum(axis=1)

    def sampleNextInternal(self, variables):
        # TODO : comment

        smplARp = variables[self.samplerEngine.I_NOISE_ARP]
        InvAutoCorrNoise = smplARp.InvAutoCorrNoise
        smplHRF = self.get_variable('hrf')
        varXh = smplHRF.varXh
        smplDrift = self.get_variable('drift')
        varMBYPl = smplDrift.varMBYPl
        smplNRL = self.get_variable('nrl')
        varNRLs = smplNRL.currentValue
        self.computeVarYTilde(varNRLs, varXh, varMBYPl)

        for i in xrange(self.nbVox):
            varYtildeTdelta = np.dot(
                self.varYTilde[:, i], InvAutoCorrNoise[:, :, i])
            self.beta[i] = 0.5 * np.dot(varYtildeTdelta, self.varYTilde[:, i])
        logger.debug('betas apost :')
        logger.debug(np.array2string(self.beta, precision=3))
        logger.debug('sigma2 ~betas/Ga(%1.3f,1)', 0.5 * (self.ny + 1))
        gammaSamples = np.random.gamma(0.5 * (self.ny + 1), 1, self.nbVox)
        self.currentValue = np.divide(self.beta, gammaSamples)
        logger.debug('All noise vars :')
        logger.debug(np.array2string(self.currentValue, precision=3))
        logger.info('noise vars = %1.3f(%1.3f)', self.currentValue.mean(),
                    self.currentValue.std())

    def finalizeSampling(self):
        del self.beta
        del self.mXhQXh
        del self.varYTilde


class NoiseARParamsSampler(xmlio.XmlInitable, GibbsSamplerVariable):
    """

    """

    P_VAL_INI = 'initialValue'
    P_SAMPLE_FLAG = 'sampleFlag'
    P_USE_TRUE_VALUE = 'useTrueValue'

    # parameters definitions and default values :
    defaultParameters = {
        P_VAL_INI: None,
        P_USE_TRUE_VALUE: False,
        P_SAMPLE_FLAG: True,
    }

    def __init__(self, do_sampling=True, use_true_value=False,
                 val_ini=None):

        # TODO : comment
        xmlio.XmlInitable.__init__(self)

        GibbsSamplerVariable.__init__(self, 'noiseARParam', valIni=val_ini,
                                      sampleFlag=do_sampling,
                                      useTrueValue=use_true_value,
                                      axes_names=['voxel'],
                                      value_label='PM AR Params')

    def linkToData(self, dataInput):
        self.dataInput = dataInput
        self.nbConditions = self.dataInput.nbConditions
        self.nbVox = self.dataInput.nbVoxels
        self.ny = self.dataInput.ny
        self.nbColX = self.dataInput.nbColX
        self.voxIdx = np.arange(self.nbVox, dtype=int)

        # Do some allocations :
        varARp = np.zeros((self.nbVox), dtype=float)

    def checkAndSetInitValue(self, variables):
        if self.useTrueValue:
            if self.trueValue is None:
                raise Exception(
                    'Needed a true value for init but None defined')
            else:
                self.currentValue = self.trueValue

        if self.currentValue is None:
            if not self.sampleFlag and self.dataInput.simulData != None \
                    and hasattr(self.dataInput.simulData.noise, 'AR'):
                if self.dataInput.simulData.noise.AR.shape[0] == 1:  # SI
                    self.currentValue = np.zeros(self.nbVox) + \
                        self.dataInput.simulData.noise.AR[0]
                else:  # SV
                    self.currentValue = self.dataInput.simulData.noise.AR[:, 0]
            else:
                self.currentValue = 0.4 * np.random.rand(self.nbVox) + .1
        # print "checkAndSetInitValue NOISE ARP"
        self.InvAutoCorrNoise = np.zeros((self.ny, self.ny, self.nbVox),
                                         dtype=float)
        self.computeInvAutoCorrNoise(self.currentValue)

    def sampleNextInternal(self, variables):
        reps = self.get_variable('noise_var').currentValue

        # TODO : precomputations for parallel AR sampling
        self.varYtilde = self.get_variable('nrl').varYtilde
        Ytilde_truncated = self.varYtilde[1:self.ny - 1, :]    #
        # compute the sequence of A_i = sum_{n=2}^{nscans -1} y_{n,i}^2, i is
        # indexing voxels
        A = np.diag(np.dot(Ytilde_truncated.transpose(), Ytilde_truncated))

        # compute the sequence of B_i = sum_{n=1}^{nscans -1} y_{n,i}y_{i,n+1},
        # i is indexing voxels
        Ytilde_shifted = self.varYtilde[:self.ny - 1, :]
        B = np.diag(np.dot(Ytilde_shifted.transpose(), self.varYtilde[1:, :]))
        M = np.divide(B, A)
        sig2 = np.divide(reps, A)
#        CommonPart = .5 * np.divide((1. - M**2),sig2)
#        CommonPartB = .5 * np.divide(2.* M - 1,log(2.)*sig2 )
#        idxM = where( M<1. )
#        Minf1 = idxM[1]
#        Msup1 = setdiff1d( self.voxIdx, Minf1 )
        # Computation of a_i = A_i/(2*reps_i)*(1-M_i^2)*(1+M_i)
#        a = zeros(self.nbVox, dtype=float)
#        a[Minf1] = multiply( CommonPart[Minf1], 1 + M[Minf1] )
#        a[Msup1] = CommonPartB[Msup1]
        # Computation of b_i = A_i/(2*reps_i)*(1-M_i^2)*(1-M_i)
#        b = zeros(self.nbVox, dtype=float)
#        b[Minf1] =  multiply( CommonPart[Minf1], 1 - M[Minf1] )
#        b[Msup1] = zeros(size(Msup1) )
        varARp = self.MH_ARsampling_gauss_proposal(sig2, M)
#	ARp = self.MH_ARsampling_optim(A,reps,M)

        self.currentValue = varARp
        logger.debug('All AR params :')
        logger.debug(np.array2string(self.currentValue, precision=3))
        logger.info('ARp = %1.3f(%1.3f)', self.currentValue.mean(),
                    self.currentValue.std())
        self.computeInvAutoCorrNoise(varARp)

    def computeInvAutoCorrNoise(self, ARp):
        logger.debug('ARp :')
        logger.debug(ARp)

        for v in xrange(self.nbVox):
            DiagAutoCorr = np.repeat(1 + ARp[v] ** 2, self.ny - 2)
            DiagAutoCorr = np.concatenate(([1], DiagAutoCorr, [1]))
            InvAutoCorrMatrix = np.diag(DiagAutoCorr, k=0)

            DiagInfAutoCorr = np.repeat(-ARp[v], self.ny - 1)
            InvAutoCorrMatrix += np.diag(DiagInfAutoCorr, k=1)
            InvAutoCorrMatrix += np.diag(DiagInfAutoCorr, k=-1)
            self.InvAutoCorrNoise[:, :, v] = InvAutoCorrMatrix

        logger.debug('InvAutoCorrNoise :')
        logger.debug(self.InvAutoCorrNoise[:, :, 0])

    def MH_ARsampling_gauss_proposal(self, sig2, M):
        ARp = np.zeros(self.nbVox, dtype=float)
        for i in xrange(self.nbVox):
            rho0 = self.currentValue[i]
            Frac_rho0 = np.sqrt(1. - rho0 ** 2)
            y = 10
            while abs(y) > 1:
                y = np.sqrt(sig2[i]) * np.random.randn(1) + M[i]
            Frac_rho = np.sqrt(1. - y ** 2)
            accept_ratio = min(
                np.array((1., Frac_rho / Frac_rho0), dtype=float))
            u = np.random.rand(1)
            if u < accept_ratio:
                ARp[i] = y
            else:
                ARp[i] = self.currentValue[i]
        return ARp

    def MH_ARsampling_optim(self, A, reps, M):
        ARp = np.zeros(self.nbVox, dtype=float)
        R = []
        for i in xrange(self.nbVox):
            rho0 = self.currentValue[i]
            Lam = A[i] / reps[i] / 2
            LamM = Lam * M[i]
            polyCoeffs = [-Lam, LamM, Lam + .5, - LamM]
            R = scipy.roots(polyCoeffs)
            x = R[abs(R) <= 1]
            # Evaluate the density function at rho0
            d = .5 + Lam * (1 + M[i] - 2 * x) * (1 + x) ^ 2
            e = .5 + Lam * (1 - M[i] + 2 * x) * (1 - x) ^ 2

            # Sampling from the right Beta distribution
            x = np.random.beta(d + 1., e + 1.)
            # change of variables
            rho = 2 * x - 1
            LogFrac = (d - .5) * log((1 + rho0) / (1 + rho)) + \
                (e - .5) * log((1 - rho0) / (1 - rho))
            LogFrac -= Lam * (rho - rho0) * (rho + rho0 - 2 * M[i])
            if LogFrac >= 0:
                accept_ratio = 1
                ARp[i] = rho
                ACCEPT = 1
            else:
                accept_ratio = np.min(array((1, np.exp(LogFrac)), dtype=float))
                u = np.random.rand(1)
                if u < accept_ratio:
                    ARp[i] = rho
                else:
                    ARp[i] = self.currentValue[i]
        return ARp

    def finalizeSampling(self):
        pass
